#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg

from senlin.conf import api
from senlin.conf import authentication
from senlin.conf import base
from senlin.conf import conductor
from senlin.conf import dispatchers
from senlin.conf import engine
from senlin.conf import health_manager
from senlin.conf import notification
from senlin.conf import receiver
from senlin.conf import revision
from senlin.conf import zaqar

CONF = cfg.CONF

api.register_opts(CONF)
authentication.register_opts(CONF)
base.register_opts(CONF)
conductor.register_opts(CONF)
dispatchers.register_opts(CONF)
engine.register_opts(CONF)
health_manager.register_opts(CONF)
notification.register_opts(CONF)
receiver.register_opts(CONF)
revision.register_opts(CONF)
zaqar.register_opts(CONF)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import time

import eventlet
from oslo_config import cfg
from oslo_context import context as oslo_context
from oslo_log import log as logging
import oslo_messaging
from osprofiler import profiler

from senlin.common import consts
from senlin.common import context
from senlin.common import messaging
from senlin.common import service
from senlin.engine.actions import base as action_mod
from senlin.engine import event as EVENT
from senlin.objects import action as ao

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

wallclock = time.time


@profiler.trace_cls("rpc")
class EngineService(service.Service):
    """RPC server for dispatching actions.

    Receive notification from dispatcher services and schedule actions.
    """

    def __init__(self, host, topic):
        super(EngineService, self).__init__(
            self.service_name, host, topic,
            threads=CONF.engine.threads
        )
        self.version = consts.RPC_API_VERSION

        self.target = None

        # TODO(Yanyan Hu): Build a DB session with full privilege
        # for DB accessing in scheduler module
        self.db_session = context.RequestContext(is_admin=True)

        # Initialize the global environment
        EVENT.load_dispatcher()

    @property
    def service_name(self):
        return 'senlin-engine'

    def start(self):
        """Start the engine.

        Note that the engine is an internal server, we are not using
        versioned object for parameter passing.
        """
        super(EngineService, self).start()

        self.target = oslo_messaging.Target(server=self.service_id,
                                            topic=self.topic,
                                            version=self.version)

        self.server = messaging.get_rpc_server(self.target, self)
        self.server.start()

    def stop(self, graceful=False):
        if self.server:
            self.server.stop()
            self.server.wait()
        super(EngineService, self).stop(graceful)

    def execute(self, func, *args, **kwargs):
        """Run the given method in a thread."""
        req_cnxt = oslo_context.get_current()
        self.tg.add_thread(
            self._start_with_trace, req_cnxt,
            self._serialize_profile_info(),
            func, *args, **kwargs
        )

    def _serialize_profile_info(self):
        prof = profiler.get()
        trace_info = None
        if prof:
            trace_info = {
                "hmac_key": prof.hmac_key,
                "base_id": prof.get_base_id(),
                "parent_id": prof.get_id()
            }
        return trace_info

    def _start_with_trace(self, cnxt, trace, func, *args, **kwargs):
        if trace:
            profiler.init(**trace)
        if cnxt is not None:
            cnxt.update_store()
        return func(*args, **kwargs)

    def listening(self, ctxt):
        """Respond affirmatively to confirm that engine is still alive."""
        return True

    def start_action(self, ctxt, action_id=None):
        """Run action(s) in sub-thread(s).

        :param action_id: ID of the action to be executed. None means all
                          ready actions will be acquired and scheduled to run.
        """
        actions_launched = 0
        max_batch_size = cfg.CONF.max_actions_per_batch
        batch_interval = cfg.CONF.batch_interval

        if action_id is not None:
            timestamp = wallclock()
            action = ao.Action.acquire(self.db_session, action_id,
                                       self.service_id,
                                       timestamp)
            if action:
                self.execute(action_mod.ActionProc, self.db_session, action.id)
                actions_launched += 1

        while True:
            timestamp = wallclock()
            action = ao.Action.acquire_first_ready(self.db_session,
                                                   self.service_id,
                                                   timestamp)
            if not action:
                break

            if max_batch_size == 0 or 'NODE' not in action.action:
                self.execute(action_mod.ActionProc, self.db_session, action.id)
                continue

            if max_batch_size > actions_launched:
                self.execute(action_mod.ActionProc, self.db_session, action.id)
                actions_launched += 1
                continue

            self.execute(action_mod.ActionProc, self.db_session, action.id)

            LOG.debug(
                'Engine %(id)s has launched %(num)s node actions '
                'consecutively, stop scheduling node action for '
                '%(interval)s second...',
                {
                    'id': self.service_id,
                    'num': max_batch_size,
                    'interval': batch_interval
                })

            sleep(batch_interval)
            actions_launched = 1

    def cancel_action(self, ctxt, action_id):
        """Cancel an action execution progress."""
        action = action_mod.Action.load(self.db_session, action_id,
                                        project_safe=False)
        action.signal(action.SIG_CANCEL)

    def suspend_action(self, ctxt, action_id):
        """Suspend an action execution progress."""
        action = action_mod.Action.load(self.db_session, action_id,
                                        project_safe=False)
        action.signal(action.SIG_SUSPEND)

    def resume_action(self, ctxt, action_id):
        """Resume an action execution progress."""
        action = action_mod.Action.load(self.db_session, action_id,
                                        project_safe=False)
        action.signal(action.SIG_RESUME)


def sleep(sleep_time):
    """Interface for sleeping."""

    eventlet.sleep(sleep_time)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlin.drivers import base
from senlin.drivers import sdk


class GlanceClient(base.DriverBase):
    """Fake Glance V2 driver."""

    def __init__(self, ctx):
        self.fake_image = {
            "created": "2015-01-01T01:02:03Z",
            "id": "70a599e0-31e7-49b7-b260-868f441e862b",
            "links": [],
            "metadata": {
                "architecture": "x86_64",
                "auto_disk_config": "True",
                "kernel_id": "nokernel",
                "ramdisk_id": "nokernel"
            },
            "minDisk": 0,
            "minRam": 0,
            "name": "cirros-0.3.5-x86_64-disk",
            "progress": 100,
            "status": "ACTIVE",
            "updated": "2011-01-01T01:02:03Z"
        }

    def image_find(self, name_or_id, ignore_missing=False):
        return sdk.FakeResourceObject(self.fake_image)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import re
import time
from unittest import mock

from oslo_config import cfg
from oslo_utils import timeutils as tu

from senlin.common import consts
from senlin.common import context
from senlin.common import exception as exc
from senlin.common import utils
from senlin.engine import health_manager as hm
from senlin.engine import node as node_mod
from senlin.engine.notifications import nova_endpoint
from senlin import objects
from senlin.objects import cluster as obj_cluster
from senlin.objects import node as obj_node
from senlin.objects import profile as obj_profile
from senlin.rpc import client as rpc_client
from senlin.tests.unit.common import base


class TestChaseUp(base.SenlinTestCase):

    def test_less_than_one_interval(self):
        start = tu.utcnow(True)
        # we assume that the delay before next line is < 5 seconds
        res = hm.chase_up(start, 5)

        self.assertLessEqual(res, 5)

    def test_more_than_one_interval(self):
        start = tu.utcnow(True)
        time.sleep(2)

        # we assume that the delay before next line is < 5 seconds
        res = hm.chase_up(start, 1)

        self.assertLessEqual(res, 1)


@mock.patch('oslo_messaging.NotificationFilter')
class TestNovaNotificationEndpoint(base.SenlinTestCase):

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_init(self, mock_rpc, mock_filter):
        x_filter = mock_filter.return_value
        event_map = {
            'compute.instance.pause.end': 'PAUSE',
            'compute.instance.power_off.end': 'POWER_OFF',
            'compute.instance.rebuild.error': 'REBUILD',
            'compute.instance.shutdown.end': 'SHUTDOWN',
            'compute.instance.soft_delete.end': 'SOFT_DELETE',
        }
        recover_action = {'operation': 'REBUILD'}
        endpoint = nova_endpoint.NovaNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )

        mock_filter.assert_called_once_with(
            publisher_id='^compute.*',
            event_type='^compute\.instance\..*',
            context={'project_id': '^PROJECT$'})
        mock_rpc.assert_called_once_with()
        self.assertEqual(x_filter, endpoint.filter_rule)
        self.assertEqual(mock_rpc.return_value, endpoint.rpc)
        for e in event_map:
            self.assertIn(e, endpoint.VM_FAILURE_EVENTS)
            self.assertEqual(event_map[e], endpoint.VM_FAILURE_EVENTS[e])
        self.assertEqual('PROJECT', endpoint.project_id)
        self.assertEqual('CLUSTER_ID', endpoint.cluster_id)

    @mock.patch.object(context.RequestContext, 'from_dict')
    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info(self, mock_rpc, mock_context, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = nova_endpoint.NovaNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {
            'metadata': {
                'cluster_id': 'CLUSTER_ID',
                'cluster_node_id': 'FAKE_NODE',
                'cluster_node_index': '123',
            },
            'instance_id': 'PHYSICAL_ID',
            'user_id': 'USER',
            'state': 'shutoff',
        }
        metadata = {'timestamp': 'TIMESTAMP'}
        call_ctx = mock.Mock()
        mock_context.return_value = call_ctx

        res = endpoint.info(ctx, 'PUBLISHER', 'compute.instance.shutdown.end',
                            payload, metadata)

        self.assertIsNone(res)
        x_rpc.call.assert_called_once_with(call_ctx, 'node_recover', mock.ANY)
        req = x_rpc.call.call_args[0][2]
        self.assertIsInstance(req, objects.NodeRecoverRequest)
        self.assertEqual('FAKE_NODE', req.identity)
        expected_params = {
            'event': 'SHUTDOWN',
            'state': 'shutoff',
            'instance_id': 'PHYSICAL_ID',
            'timestamp': 'TIMESTAMP',
            'publisher': 'PUBLISHER',
            'operation': 'REBUILD'
        }
        self.assertEqual(expected_params, req.params)

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_no_metadata(self, mock_rpc, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = nova_endpoint.NovaNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {'metadata': {}}
        metadata = {'timestamp': 'TIMESTAMP'}

        res = endpoint.info(ctx, 'PUBLISHER', 'compute.instance.delete.end',
                            payload, metadata)

        self.assertIsNone(res)
        self.assertEqual(0, x_rpc.node_recover.call_count)

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_no_cluster_in_metadata(self, mock_rpc, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = nova_endpoint.NovaNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {'metadata': {'foo': 'bar'}}
        metadata = {'timestamp': 'TIMESTAMP'}

        res = endpoint.info(ctx, 'PUBLISHER', 'compute.instance.delete.end',
                            payload, metadata)

        self.assertIsNone(res)
        self.assertEqual(0, x_rpc.node_recover.call_count)

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_cluster_id_not_match(self, mock_rpc, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = nova_endpoint.NovaNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {'metadata': {'cluster_id': 'FOOBAR'}}
        metadata = {'timestamp': 'TIMESTAMP'}

        res = endpoint.info(ctx, 'PUBLISHER', 'compute.instance.delete.end',
                            payload, metadata)

        self.assertIsNone(res)
        self.assertEqual(0, x_rpc.node_recover.call_count)

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_event_type_not_interested(self, mock_rpc, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = nova_endpoint.NovaNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {'metadata': {'cluster_id': 'CLUSTER_ID'}}
        metadata = {'timestamp': 'TIMESTAMP'}

        res = endpoint.info(ctx, 'PUBLISHER', 'compute.instance.delete.start',
                            payload, metadata)

        self.assertIsNone(res)
        self.assertEqual(0, x_rpc.node_recover.call_count)

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_no_node_id(self, mock_rpc, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = nova_endpoint.NovaNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {'metadata': {'cluster_id': 'CLUSTER_ID'}}
        metadata = {'timestamp': 'TIMESTAMP'}

        res = endpoint.info(ctx, 'PUBLISHER', 'compute.instance.delete.end',
                            payload, metadata)

        self.assertIsNone(res)
        self.assertEqual(0, x_rpc.node_recover.call_count)

    @mock.patch.object(context.RequestContext, 'from_dict')
    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_default_values(self, mock_rpc, mock_context, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = nova_endpoint.NovaNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {
            'metadata': {
                'cluster_id': 'CLUSTER_ID',
                'cluster_node_id': 'NODE_ID'
            },
            'user_id': 'USER',
        }
        metadata = {'timestamp': 'TIMESTAMP'}
        call_ctx = mock.Mock()
        mock_context.return_value = call_ctx

        res = endpoint.info(ctx, 'PUBLISHER', 'compute.instance.shutdown.end',
                            payload, metadata)

        self.assertIsNone(res)
        x_rpc.call.assert_called_once_with(call_ctx, 'node_recover', mock.ANY)
        req = x_rpc.call.call_args[0][2]
        self.assertIsInstance(req, objects.NodeRecoverRequest)
        self.assertEqual('NODE_ID', req.identity)
        expected_params = {
            'event': 'SHUTDOWN',
            'state': 'Unknown',
            'instance_id': 'Unknown',
            'timestamp': 'TIMESTAMP',
            'publisher': 'PUBLISHER',
            'operation': 'REBUILD',
        }
        self.assertEqual(expected_params, req.params)


@mock.patch(
    'senlin.engine.notifications.heat_endpoint.HeatNotificationEndpoint')
@mock.patch(
    'senlin.engine.notifications.nova_endpoint.NovaNotificationEndpoint')
@mock.patch('oslo_messaging.get_notification_transport')
@mock.patch('oslo_messaging.get_notification_listener')
class TestListenerProc(base.SenlinTestCase):

    def test_listener_proc_nova(self, mock_listener, mock_transport,
                                mock_novaendpoint, mock_heatendpoint):
        cfg.CONF.set_override('nova_control_exchange', 'FAKE_EXCHANGE',
                              group='health_manager')

        x_listener = mock.Mock()
        mock_listener.return_value = x_listener
        x_transport = mock.Mock()
        mock_transport.return_value = x_transport
        x_endpoint = mock.Mock()
        mock_novaendpoint.return_value = x_endpoint

        recover_action = {'operation': 'REBUILD'}
        res = hm.ListenerProc('FAKE_EXCHANGE', 'PROJECT_ID', 'CLUSTER_ID',
                              recover_action)

        self.assertIsNone(res)
        mock_transport.assert_called_once_with(cfg.CONF)
        mock_novaendpoint.assert_called_once_with('PROJECT_ID', 'CLUSTER_ID',
                                                  recover_action)
        mock_listener.assert_called_once_with(
            x_transport, [mock_novaendpoint().target], [x_endpoint],
            executor='threading', pool="senlin-listeners")
        x_listener.start.assert_called_once_with()

    def test_listener_proc_heat(self, mock_listener, mock_transport,
                                mock_novaendpoint, mock_heatendpoint):
        x_listener = mock.Mock()
        mock_listener.return_value = x_listener
        x_transport = mock.Mock()
        mock_transport.return_value = x_transport
        x_endpoint = mock.Mock()
        mock_heatendpoint.return_value = x_endpoint

        recover_action = {'operation': 'REBUILD'}
        res = hm.ListenerProc('heat', 'PROJECT_ID', 'CLUSTER_ID',
                              recover_action)

        self.assertIsNone(res)
        mock_transport.assert_called_once_with(cfg.CONF)
        mock_heatendpoint.assert_called_once_with('PROJECT_ID', 'CLUSTER_ID',
                                                  recover_action)
        mock_listener.assert_called_once_with(
            x_transport, [mock_heatendpoint().target], [x_endpoint],
            executor='threading', pool="senlin-listeners")
        x_listener.start.assert_called_once_with()


class TestHealthCheckType(base.SenlinTestCase):
    def setUp(self):
        super(TestHealthCheckType, self).setUp()

        self.hc = hm.NodePollStatusHealthCheck(
            cluster_id='CLUSTER_ID', interval=1, node_update_timeout=1,
            params=''
        )

    def test_factory(self):
        cid = 'CLUSTER_ID'
        interval = 1
        params = {
            'detection_modes': [
                {
                    'type': 'NODE_STATUS_POLLING',
                    'poll_url': '',
                    'poll_url_ssl_verify': True,
                    'poll_url_conn_error_as_unhealthy': True,
                    'poll_url_healthy_response': '',
                    'poll_url_retry_limit': '',
                    'poll_url_retry_interval': ''
                },
                {
                    'type': 'HYPERVISOR_STATUS_POLLING',
                    'poll_url': '',
                    'poll_url_ssl_verify': True,
                    'poll_url_conn_error_as_unhealthy': True,
                    'poll_url_healthy_response': '',
                    'poll_url_retry_limit': '',
                    'poll_url_retry_interval': ''
                },
                {
                    'type': 'NODE_STATUS_POLL_URL',
                    'poll_url': '',
                    'poll_url_ssl_verify': True,
                    'poll_url_conn_error_as_unhealthy': True,
                    'poll_url_healthy_response': '',
                    'poll_url_retry_limit': '',
                    'poll_url_retry_interval': ''
                }
            ],
            'node_update_timeout': 300,
        }

        for d in params['detection_modes']:
            hc = hm.HealthCheckType.factory(d['type'], cid, interval, params)

            self.assertEqual(cid, hc.cluster_id)
            self.assertEqual(interval, hc.interval)
            self.assertEqual(d, hc.params)
            self.assertEqual(
                params['node_update_timeout'], hc.node_update_timeout)

    def test_factory_invalid_type(self):
        cid = 'CLUSTER_ID'
        interval = 1
        params = {
            'detection_modes': [
                {
                    'type': 'blah',
                    'poll_url': '',
                    'poll_url_ssl_verify': True,
                    'poll_url_conn_error_as_unhealthy': True,
                    'poll_url_healthy_response': '',
                    'poll_url_retry_limit': '',
                    'poll_url_retry_interval': ''
                },
            ],
            'node_update_timeout': 300,
        }

        with self.assertRaisesRegex(Exception, 'Invalid detection type: blah'):
            hm.HealthCheckType.factory('blah', cid, interval, params)

    def test_factory_same_type_twice(self):
        cid = 'CLUSTER_ID'
        interval = 1
        params = {
            'detection_modes': [
                {
                    'type': 'NODE_STATUS_POLLING',
                    'poll_url': '',
                    'poll_url_ssl_verify': True,
                    'poll_url_conn_error_as_unhealthy': True,
                    'poll_url_healthy_response': '',
                    'poll_url_retry_limit': '',
                    'poll_url_retry_interval': ''
                },
                {
                    'type': 'NODE_STATUS_POLLING',
                    'poll_url': '',
                    'poll_url_ssl_verify': True,
                    'poll_url_conn_error_as_unhealthy': True,
                    'poll_url_healthy_response': '',
                    'poll_url_retry_limit': '',
                    'poll_url_retry_interval': ''
                }
            ],
            'node_update_timeout': 300,
        }

        with self.assertRaisesRegex(
                Exception,
                '.*Encountered 2 instances of type NODE_STATUS_POLLING'):
            hm.HealthCheckType.factory(
                'NODE_STATUS_POLLING', cid, interval, params)


class TestNodePollStatusHealthCheck(base.SenlinTestCase):
    def setUp(self):
        super(TestNodePollStatusHealthCheck, self).setUp()

        self.hc = hm.NodePollStatusHealthCheck(
            cluster_id='CLUSTER_ID',
            interval=1, node_update_timeout=1, params=''
        )

    @mock.patch.object(node_mod.Node, '_from_object')
    @mock.patch.object(tu, 'is_older_than')
    def test_run_health_check_healthy(self, mock_tu, mock_node_obj):
        x_entity = mock.Mock()
        x_entity.do_healthcheck.return_value = True
        mock_node_obj.return_value = x_entity

        ctx = mock.Mock()
        node = mock.Mock(id='FAKE_NODE1', status="ERROR",
                         updated_at='2018-08-13 18:00:00',
                         init_at='2018-08-13 17:00:00')

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_tu.assert_not_called()

    @mock.patch.object(node_mod.Node, '_from_object')
    @mock.patch.object(tu, 'is_older_than')
    def test_run_health_check_healthy_internal_error(
            self, mock_tu, mock_node_obj):
        x_entity = mock.Mock()
        x_entity.do_healthcheck.side_effect = exc.InternalError(
            message='error')
        mock_node_obj.return_value = x_entity

        ctx = mock.Mock()
        node = mock.Mock(id='FAKE_NODE1', status="ERROR",
                         updated_at='2018-08-13 18:00:00',
                         init_at='2018-08-13 17:00:00')

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_tu.assert_not_called()

    @mock.patch.object(node_mod.Node, '_from_object')
    @mock.patch.object(tu, 'is_older_than')
    def test_run_health_check_unhealthy(self, mock_tu, mock_node_obj):
        x_entity = mock.Mock()
        x_entity.do_healthcheck.return_value = False
        mock_node_obj.return_value = x_entity

        mock_tu.return_value = True

        ctx = mock.Mock()
        node = mock.Mock(id='FAKE_NODE1', status="ERROR",
                         updated_at='2018-08-13 18:00:00',
                         init_at='2018-08-13 17:00:00')

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertFalse(res)
        mock_tu.assert_called_once_with(node.updated_at, 1)

    @mock.patch.object(node_mod.Node, '_from_object')
    @mock.patch.object(tu, 'is_older_than')
    def test_run_health_check_unhealthy_within_timeout(
            self, mock_tu, mock_node_obj):
        x_entity = mock.Mock()
        x_entity.do_healthcheck.return_value = False
        mock_node_obj.return_value = x_entity

        mock_tu.return_value = False

        ctx = mock.Mock()
        node = mock.Mock(id='FAKE_NODE1', status="ERROR",
                         updated_at='2018-08-13 18:00:00',
                         init_at='2018-08-13 17:00:00')

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_tu.assert_called_once_with(node.updated_at, 1)


class TestHypervisorPollStatusHealthCheck(base.SenlinTestCase):
    def setUp(self):
        super(TestHypervisorPollStatusHealthCheck, self).setUp()

        self.hc = hm.HypervisorPollStatusHealthCheck(
            cluster_id='CLUSTER_ID',
            interval=1, node_update_timeout=1, params=''
        )

    @mock.patch.object(node_mod.Node, '_from_object')
    @mock.patch.object(tu, 'is_older_than')
    def test_run_health_check_healthy(self, mock_tu, mock_node_obj):
        x_entity = mock.Mock()
        x_entity.do_healthcheck.return_value = True
        mock_node_obj.return_value = x_entity

        ctx = mock.Mock()
        node = mock.Mock(id='FAKE_NODE1', status="ERROR",
                         updated_at='2018-08-13 18:00:00',
                         init_at='2018-08-13 17:00:00')

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_tu.assert_not_called()

    @mock.patch.object(node_mod.Node, '_from_object')
    @mock.patch.object(tu, 'is_older_than')
    def test_run_health_check_healthy_internal_error(
            self, mock_tu, mock_node_obj):
        x_entity = mock.Mock()
        x_entity.do_healthcheck.side_effect = exc.InternalError(
            message='error')
        mock_node_obj.return_value = x_entity

        ctx = mock.Mock()
        node = mock.Mock(id='FAKE_NODE1', status="ERROR",
                         updated_at='2018-08-13 18:00:00',
                         init_at='2018-08-13 17:00:00')

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_tu.assert_not_called()

    @mock.patch.object(node_mod.Node, '_from_object')
    @mock.patch.object(tu, 'is_older_than')
    def test_run_health_check_unhealthy(self, mock_tu, mock_node_obj):
        x_entity = mock.Mock()
        x_entity.do_healthcheck.return_value = False
        mock_node_obj.return_value = x_entity

        mock_tu.return_value = True

        ctx = mock.Mock()
        node = mock.Mock(id='FAKE_NODE1', status="ERROR",
                         updated_at='2018-08-13 18:00:00',
                         init_at='2018-08-13 17:00:00')

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertFalse(res)
        mock_tu.assert_called_once_with(node.updated_at, 1)

    @mock.patch.object(node_mod.Node, '_from_object')
    @mock.patch.object(tu, 'is_older_than')
    def test_run_health_check_unhealthy_within_timeout(
            self, mock_tu, mock_node_obj):
        x_entity = mock.Mock()
        x_entity.do_healthcheck.return_value = False
        mock_node_obj.return_value = x_entity

        mock_tu.return_value = False

        ctx = mock.Mock()
        node = mock.Mock(id='FAKE_NODE1', status="ERROR",
                         updated_at='2018-08-13 18:00:00',
                         init_at='2018-08-13 17:00:00')

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_tu.assert_called_once_with(node.updated_at, 1)


class TestNodePollUrlHealthCheck(base.SenlinTestCase):
    def setUp(self):
        super(TestNodePollUrlHealthCheck, self).setUp()

        default_params = {
            'poll_url': 'FAKE_POLL_URL',
            'poll_url_ssl_verify': True,
            'poll_url_conn_error_as_unhealthy': True,
            'poll_url_healthy_response': 'FAKE_HEALTHY_PATTERN',
            'poll_url_retry_limit': 2,
            'poll_url_retry_interval': 1,
            'node_update_timeout': 5
        }

        self.hc = hm.NodePollUrlHealthCheck(
            cluster_id='CLUSTER_ID', interval=1, node_update_timeout=1,
            params=default_params
        )

    def test_expand_url_template(self):
        url_template = 'https://abc123/foo/bar'
        node = mock.Mock()

        # do it
        res = self.hc._expand_url_template(url_template, node)

        self.assertEqual(res, url_template)

    def test_expand_url_template_nodename(self):
        node = mock.Mock()
        node.name = 'name'
        url_template = 'https://abc123/{nodename}/bar'
        expanded_url = 'https://abc123/{}/bar'.format(node.name)

        # do it
        res = self.hc._expand_url_template(url_template, node)

        self.assertEqual(res, expanded_url)

    @mock.patch.object(tu, "is_older_than")
    @mock.patch.object(hm.NodePollUrlHealthCheck, "_expand_url_template")
    @mock.patch.object(utils, 'url_fetch')
    def test_run_health_check_healthy(
            self, mock_url_fetch, mock_expand_url, mock_time):
        ctx = mock.Mock()
        node = mock.Mock()
        node.status = consts.NS_ACTIVE
        mock_time.return_value = True
        mock_expand_url.return_value = 'FAKE_EXPANDED_URL'
        mock_url_fetch.return_value = ("Healthy because this return value "
                                       "contains FAKE_HEALTHY_PATTERN")

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_url_fetch.assert_called_once_with('FAKE_EXPANDED_URL', timeout=1,
                                               verify=True)

    @mock.patch.object(tu, "is_older_than")
    @mock.patch.object(hm.NodePollUrlHealthCheck, "_expand_url_template")
    @mock.patch.object(utils, 'url_fetch')
    def test_run_health_check_healthy_min_timeout(
            self, mock_url_fetch, mock_expand_url, mock_time):
        ctx = mock.Mock()
        node = mock.Mock()
        node.status = consts.NS_ACTIVE
        mock_time.return_value = True
        mock_expand_url.return_value = 'FAKE_EXPANDED_URL'
        mock_url_fetch.return_value = ("Healthy because this return value "
                                       "contains FAKE_HEALTHY_PATTERN")

        self.hc.params['poll_url_retry_interval'] = 0

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_url_fetch.assert_called_once_with('FAKE_EXPANDED_URL', timeout=1,
                                               verify=True)

    @mock.patch.object(tu, "is_older_than")
    @mock.patch.object(hm.NodePollUrlHealthCheck, "_expand_url_template")
    @mock.patch.object(utils, 'url_fetch')
    def test_run_health_check_healthy_timeout(
            self, mock_url_fetch, mock_expand_url, mock_time):
        ctx = mock.Mock()
        node = mock.Mock()
        node.status = consts.NS_ACTIVE
        mock_time.return_value = True
        mock_expand_url.return_value = 'FAKE_EXPANDED_URL'
        mock_url_fetch.return_value = ("Healthy because this return value "
                                       "contains FAKE_HEALTHY_PATTERN")

        self.hc.params['poll_url_retry_interval'] = 100

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_url_fetch.assert_called_once_with('FAKE_EXPANDED_URL', timeout=10,
                                               verify=True)

    @mock.patch.object(tu, "is_older_than")
    @mock.patch.object(hm.NodePollUrlHealthCheck, "_expand_url_template")
    @mock.patch.object(utils, 'url_fetch')
    def test_run_health_check_unhealthy_inactive(
            self, mock_url_fetch, mock_expand_url, mock_time):
        ctx = mock.Mock()
        node = mock.Mock()
        node.status = consts.NS_RECOVERING
        mock_time.return_value = True
        mock_expand_url.return_value = 'FAKE_EXPANDED_URL'
        mock_url_fetch.return_value = ""

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_url_fetch.assert_not_called()

    @mock.patch.object(tu, "is_older_than")
    @mock.patch.object(hm.NodePollUrlHealthCheck, "_expand_url_template")
    @mock.patch.object(utils, 'url_fetch')
    def test_run_health_check_unhealthy_update_timeout(
            self, mock_url_fetch, mock_expand_url, mock_time):
        ctx = mock.Mock()
        node = mock.Mock()
        node.id = 'FAKE_NODE_ID'
        node.updated_at = 'FAKE_UPDATE_TIME'
        node.status = consts.NS_ACTIVE
        mock_time.return_value = False
        mock_expand_url.return_value = 'FAKE_EXPANDED_URL'
        mock_url_fetch.return_value = ""

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_url_fetch.assert_has_calls(
            [mock.call('FAKE_EXPANDED_URL', timeout=1, verify=True)])

    @mock.patch.object(tu, "is_older_than")
    @mock.patch.object(hm.NodePollUrlHealthCheck, "_expand_url_template")
    @mock.patch.object(utils, 'url_fetch')
    def test_run_health_check_unhealthy_init_timeout(
            self, mock_url_fetch, mock_expand_url, mock_time):
        ctx = mock.Mock()
        node = mock.Mock()
        node.id = 'FAKE_NODE_ID'
        node.updated_at = None
        node.init_at = 'FAKE_INIT_TIME'
        node.status = consts.NS_ACTIVE
        mock_time.return_value = False
        mock_expand_url.return_value = 'FAKE_EXPANDED_URL'
        mock_url_fetch.return_value = ""

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_url_fetch.assert_has_calls(
            [mock.call('FAKE_EXPANDED_URL', timeout=1, verify=True)])

    @mock.patch.object(tu, "is_older_than")
    @mock.patch.object(hm.NodePollUrlHealthCheck, "_expand_url_template")
    @mock.patch.object(utils, 'url_fetch')
    def test_run_health_check_unhealthy(self, mock_url_fetch, mock_expand_url,
                                        mock_time):
        ctx = mock.Mock()
        node = mock.Mock()
        node.status = consts.NS_ACTIVE
        node.id = 'FAKE_ID'
        mock_time.return_value = True
        mock_expand_url.return_value = 'FAKE_EXPANDED_URL'
        mock_url_fetch.return_value = ""

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertFalse(res)
        mock_url_fetch.assert_has_calls(
            [
                mock.call('FAKE_EXPANDED_URL', timeout=1, verify=True),
                mock.call('FAKE_EXPANDED_URL', timeout=1, verify=True)
            ]
        )

    @mock.patch.object(tu, "is_older_than")
    @mock.patch.object(hm.NodePollUrlHealthCheck, "_expand_url_template")
    @mock.patch.object(utils, 'url_fetch')
    def test_run_health_check_conn_error(self,
                                         mock_url_fetch,
                                         mock_expand_url, mock_time):
        ctx = mock.Mock()
        node = mock.Mock()
        node.status = consts.NS_ACTIVE
        node.id = 'FAKE_ID'
        mock_time.return_value = True
        mock_expand_url.return_value = 'FAKE_EXPANDED_URL'
        mock_url_fetch.side_effect = utils.URLFetchError("Error")

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertFalse(res)
        mock_url_fetch.assert_has_calls(
            [
                mock.call('FAKE_EXPANDED_URL', timeout=1, verify=True),
                mock.call('FAKE_EXPANDED_URL', timeout=1, verify=True)
            ]
        )

    @mock.patch.object(tu, "is_older_than")
    @mock.patch.object(hm.NodePollUrlHealthCheck, "_expand_url_template")
    @mock.patch.object(utils, 'url_fetch')
    def test_run_health_check_conn_other_error(self,
                                               mock_url_fetch,
                                               mock_expand_url, mock_time):
        ctx = mock.Mock()
        node = mock.Mock()
        node.status = consts.NS_ACTIVE
        node.id = 'FAKE_ID'
        mock_time.return_value = True
        mock_expand_url.side_effect = Exception('blah')

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_url_fetch.assert_not_called()

    @mock.patch.object(tu, "is_older_than")
    @mock.patch.object(hm.NodePollUrlHealthCheck, "_expand_url_template")
    @mock.patch.object(utils, 'url_fetch')
    def test_run_health_check_conn_error_noop(
            self, mock_url_fetch, mock_expand_url, mock_time):
        ctx = mock.Mock()
        node = mock.Mock()
        node.status = consts.NS_ACTIVE
        node.id = 'FAKE_ID'
        mock_time.return_value = True
        mock_expand_url.return_value = 'FAKE_EXPANDED_URL'
        mock_url_fetch.side_effect = utils.URLFetchError("Error")

        self.hc.params['poll_url_conn_error_as_unhealthy'] = False

        # do it
        res = self.hc.run_health_check(ctx, node)

        self.assertTrue(res)
        mock_url_fetch.assert_has_calls(
            [
                mock.call('FAKE_EXPANDED_URL', timeout=1, verify=True),
            ]
        )


class TestHealthCheck(base.SenlinTestCase):

    def setUp(self):
        super(TestHealthCheck, self).setUp()
        ctx = mock.Mock()
        self.fake_rpc = mock.Mock()
        with mock.patch.object(rpc_client, 'get_engine_client',
                               return_value=self.fake_rpc):
            self.hc = hm.HealthCheck(
                ctx=ctx,
                engine_id='ENGINE_ID',
                cluster_id='CID',
                check_type=consts.NODE_STATUS_POLLING,
                interval=60,
                node_update_timeout=60,
                params={
                    'node_update_timeout': 60,
                    'detection_modes': [
                        {'type': consts.NODE_STATUS_POLLING}
                    ],
                    'recovery_conditional': consts.ANY_FAILED
                },
                enabled=True)

    def test_get_health_check_types_polling(self):
        self.hc.get_health_check_types()
        self.assertEqual(consts.POLLING, self.hc.type)

    def test_get_health_check_types_events(self):
        self.hc.check_type = consts.LIFECYCLE_EVENTS
        self.hc.get_health_check_types()
        self.assertEqual(consts.EVENTS, self.hc.type)

    def test_get_recover_actions(self):
        self.hc.params = {
            'node_delete_timeout': 60,
            'node_force_recreate': True,
            'recover_action': [{'name': 'FAKE_RECOVER_ACTION'}]
        }
        self.hc.get_recover_actions()
        self.assertEqual(self.hc.params['node_delete_timeout'],
                         self.hc.recover_action['delete_timeout'])
        self.assertEqual(self.hc.params['node_force_recreate'],
                         self.hc.recover_action['force_recreate'])
        self.assertEqual(self.hc.params['recover_action'][0]['name'],
                         self.hc.recover_action['operation'])

    @mock.patch.object(obj_node.Node, 'get_all_by_cluster')
    @mock.patch.object(hm.HealthCheck, "_recover_node")
    @mock.patch.object(hm.HealthCheck, "_wait_for_action")
    @mock.patch.object(obj_cluster.Cluster, 'get')
    @mock.patch.object(context, 'get_service_context')
    def test_execute_health_check_any_mode_healthy(
            self, mock_ctx, mock_get, mock_wait, mock_recover, mock_nodes):
        x_cluster = mock.Mock(user='USER_ID', project='PROJECT_ID',
                              id='CID')
        mock_get.return_value = x_cluster

        ctx = mock.Mock()
        mock_ctx.return_value = ctx

        mock_wait.return_value = (True, "")

        x_node1 = mock.Mock(id='FAKE_NODE1', status="ERROR")
        x_node2 = mock.Mock(id='FAKE_NODE2', status="ERROR")
        mock_nodes.return_value = [x_node1, x_node2]

        hc_true = {'run_health_check.return_value': True}

        hc_test_values = [
            [
                mock.Mock(**hc_true),
                mock.Mock(**hc_true),
                mock.Mock(**hc_true),
            ],
        ]

        for hc_mocks in hc_test_values:
            self.hc.health_check_types = hc_mocks

            mock_get.reset_mock()
            mock_ctx.reset_mock()
            mock_recover.reset_mock()
            mock_wait.reset_mock()

            # do it
            self.hc.execute_health_check()

            mock_get.assert_called_once_with(self.hc.ctx, 'CID',
                                             project_safe=False)
            mock_ctx.assert_called_once_with(user_id=x_cluster.user,
                                             project_id=x_cluster.project)

            for mock_hc in hc_mocks:
                mock_hc.run_health_check.assert_has_calls(
                    [
                        mock.call(ctx, x_node1),
                        mock.call(ctx, x_node2)
                    ]
                )

            mock_recover.assert_not_called()
            mock_wait.assert_not_called()

    @mock.patch.object(obj_node.Node, 'get_all_by_cluster')
    @mock.patch.object(hm.HealthCheck, "_recover_node")
    @mock.patch.object(hm.HealthCheck, "_wait_for_action")
    @mock.patch.object(obj_cluster.Cluster, 'get')
    @mock.patch.object(context, 'get_service_context')
    def test_execute_health_check_all_mode_unhealthy(
            self, mock_ctx, mock_get, mock_wait, mock_recover, mock_nodes):
        self.hc.cluster_id = 'CLUSTER_ID'
        self.hc.interval = 1
        self.hc.recovery_cond = consts.ALL_FAILED
        self.hc.node_update_timeout = 1
        self.hc.recovery_action = {'operation': 'REBUILD'}

        x_cluster = mock.Mock(user='USER_ID', project='PROJECT_ID',
                              id='CLUSTER_ID')
        mock_get.return_value = x_cluster

        ctx = mock.Mock()
        mock_ctx.return_value = ctx

        mock_wait.return_value = (True, "")

        x_node = mock.Mock(id='FAKE_NODE', status="ERROR")
        mock_nodes.return_value = [x_node]

        mock_recover.return_value = {'action': 'FAKE_ACTION_ID'}

        hc_false = {'run_health_check.return_value': False}

        hc_test_values = [
            [
                mock.Mock(**hc_false),
            ]
        ]

        for hc_mocks in hc_test_values:
            self.hc.health_check_types = hc_mocks

            mock_get.reset_mock()
            mock_ctx.reset_mock()
            mock_recover.reset_mock()
            mock_wait.reset_mock()

            # do it
            self.hc.execute_health_check()

            mock_get.assert_called_once_with(self.hc.ctx, 'CLUSTER_ID',
                                             project_safe=False)
            mock_ctx.assert_called_once_with(user_id=x_cluster.user,
                                             project_id=x_cluster.project)

            for mock_hc in hc_mocks:
                mock_hc.run_health_check.assert_has_calls(
                    [
                        mock.call(ctx, x_node)
                    ]
                )

            mock_recover.assert_called_once_with(ctx, 'FAKE_NODE')
            mock_wait.assert_called_once_with(
                ctx, 'FAKE_ACTION_ID', self.hc.node_update_timeout)

    @mock.patch.object(obj_cluster.Cluster, 'get')
    @mock.patch.object(context, 'get_service_context')
    def test_execute_health_check_cluster_not_found(self, mock_ctx, mock_get):
        mock_get.return_value = None

        self.hc.execute_health_check()

        mock_ctx.assert_not_called()

    @mock.patch.object(hm.HealthCheck, "_recover_node")
    def test_check_node_health_any_failed(self, mock_recover):
        x_cluster = mock.Mock(user='USER_ID', project='PROJECT_ID',
                              id='CLUSTER_ID')
        x_node = mock.Mock(id='FAKE_NODE', status="ERROR")
        ctx = mock.Mock()

        self.hc.params['recovery_conditional'] = consts.ANY_FAILED
        mock_hc_1 = mock.Mock()
        mock_hc_1.run_health_check.return_value = True
        mock_hc_2 = mock.Mock()
        mock_hc_2.run_health_check.return_value = False

        self.hc.health_check_types = [mock_hc_1, mock_hc_2]

        self.hc._check_node_health(ctx, x_node, x_cluster)

        mock_hc_1.run_health_check.assert_called_once_with(ctx, x_node)
        mock_hc_2.run_health_check.assert_called_once_with(ctx, x_node)
        mock_recover.assert_called_once_with(ctx, x_node.id)

    @mock.patch.object(hm.HealthCheck, "_recover_node")
    def test_check_node_health_all_failed(self, mock_recover):
        x_cluster = mock.Mock(user='USER_ID', project='PROJECT_ID',
                              id='CLUSTER_ID')
        x_node = mock.Mock(id='FAKE_NODE', status="ERROR")
        ctx = mock.Mock()

        self.hc.params['recovery_conditional'] = consts.ALL_FAILED
        mock_hc_1 = mock.Mock()
        mock_hc_1.run_health_check.return_value = False
        mock_hc_2 = mock.Mock()
        mock_hc_2.run_health_check.return_value = False

        self.hc.health_check_types = [mock_hc_1, mock_hc_2]

        self.hc._check_node_health(ctx, x_node, x_cluster)

        mock_hc_1.run_health_check.assert_called_once_with(ctx, x_node)
        mock_hc_2.run_health_check.assert_called_once_with(ctx, x_node)
        mock_recover.assert_called_once_with(ctx, x_node.id)

    @mock.patch.object(hm.HealthCheck, "_recover_node")
    def test_check_node_health_all_failed_negative(self, mock_recover):
        x_cluster = mock.Mock(user='USER_ID', project='PROJECT_ID',
                              id='CLUSTER_ID')
        x_node = mock.Mock(id='FAKE_NODE', status="ERROR")
        ctx = mock.Mock()

        self.hc.params['recovery_conditional'] = consts.ALL_FAILED
        mock_hc_1 = mock.Mock()
        mock_hc_1.run_health_check.return_value = False
        mock_hc_2 = mock.Mock()
        mock_hc_2.run_health_check.return_value = True

        self.hc.health_check_types = [mock_hc_1, mock_hc_2]

        self.hc._check_node_health(ctx, x_node, x_cluster)

        mock_hc_1.run_health_check.assert_called_once_with(ctx, x_node)
        mock_hc_2.run_health_check.assert_called_once_with(ctx, x_node)
        mock_recover.assert_not_called()

    @mock.patch('senlin.objects.ActionGetRequest')
    def test_wait_for_action(self, mock_action_req):
        x_req = mock.Mock()
        mock_action_req.return_value = x_req

        x_action = {'status': consts.ACTION_SUCCEEDED}
        self.fake_rpc.call.return_value = x_action

        ctx = mock.Mock()
        action_id = 'FAKE_ACTION_ID'
        timeout = 5

        # do it
        res, err = self.hc._wait_for_action(ctx, action_id, timeout)

        self.assertTrue(res)
        self.assertEqual(err, '')
        self.fake_rpc.call.assert_called_with(ctx, 'action_get', x_req)

    @mock.patch('senlin.objects.ActionGetRequest')
    def test_wait_for_action_success_before_timeout(self, mock_action_req):
        x_req = mock.Mock()
        mock_action_req.return_value = x_req

        x_action1 = {'status': consts.ACTION_RUNNING}
        x_action2 = {'status': consts.ACTION_SUCCEEDED}
        self.fake_rpc.call.side_effect = [x_action1, x_action2]

        ctx = mock.Mock()
        action_id = 'FAKE_ACTION_ID'
        timeout = 5

        # do it
        res, err = self.hc._wait_for_action(ctx, action_id, timeout)

        self.assertTrue(res)
        self.assertEqual(err, '')
        self.fake_rpc.call.assert_has_calls(
            [
                mock.call(ctx, 'action_get', x_req),
                mock.call(ctx, 'action_get', x_req)
            ]
        )

    @mock.patch('senlin.objects.ActionGetRequest')
    def test_wait_for_action_timeout(self, mock_action_req):
        x_req = mock.Mock()
        mock_action_req.return_value = x_req

        x_action = {'status': consts.ACTION_RUNNING}
        self.fake_rpc.call.return_value = x_action

        ctx = mock.Mock()
        action_id = 'FAKE_ACTION_ID'
        timeout = 5

        # do it
        res, err = self.hc._wait_for_action(ctx, action_id, timeout)

        self.assertFalse(res)
        self.assertTrue(re.search('timeout', err, re.IGNORECASE))
        self.fake_rpc.call.assert_has_calls(
            [
                mock.call(ctx, 'action_get', x_req)
            ]
        )

    @mock.patch('senlin.objects.ActionGetRequest')
    def test_wait_for_action_failed(self, mock_action_req):
        x_req = mock.Mock()
        mock_action_req.return_value = x_req

        x_action = {'status': consts.ACTION_FAILED}
        self.fake_rpc.call.return_value = x_action

        ctx = mock.Mock()
        action_id = 'FAKE_ACTION_ID'
        timeout = 5

        # do it
        res, err = self.hc._wait_for_action(ctx, action_id, timeout)

        self.assertFalse(res)
        self.assertEqual(err, 'Cluster check action failed or cancelled')
        self.fake_rpc.call.assert_called_with(ctx, 'action_get', x_req)

    @mock.patch('senlin.objects.ActionGetRequest')
    def test_wait_for_action_cancelled(self, mock_action_req):
        x_req = mock.Mock()
        mock_action_req.return_value = x_req

        x_action = {'status': consts.ACTION_CANCELLED}
        self.fake_rpc.call.return_value = x_action

        ctx = mock.Mock()
        action_id = 'FAKE_ACTION_ID'
        timeout = 5

        # do it
        res, err = self.hc._wait_for_action(ctx, action_id, timeout)

        self.assertFalse(res)
        self.assertEqual(err, 'Cluster check action failed or cancelled')
        self.fake_rpc.call.assert_called_with(ctx, 'action_get', x_req)

    @mock.patch('senlin.objects.NodeRecoverRequest', autospec=True)
    def test_recover_node(self, mock_req):
        ctx = mock.Mock()
        node_id = 'FAKE_NODE'
        self.hc.recover_action = {'operation': 'REBUILD'}

        x_req = mock.Mock
        mock_req.return_value = x_req

        x_action = {'action': 'RECOVER_ID1'}
        self.fake_rpc.call.return_value = x_action

        # do it
        res = self.hc._recover_node(ctx, node_id)

        self.assertEqual(x_action, res)
        mock_req.assert_called_once_with(
            identity=node_id, params=self.hc.recover_action)
        self.fake_rpc.call.assert_called_once_with(ctx, 'node_recover', x_req)

    @mock.patch('senlin.objects.NodeRecoverRequest', autospec=True)
    def test_recover_node_failed(self, mock_req):
        ctx = mock.Mock()
        node_id = 'FAKE_NODE'
        self.hc.recover_action = {'operation': 'REBUILD'}

        x_req = mock.Mock
        mock_req.return_value = x_req

        self.fake_rpc.call.side_effect = Exception('boom')

        # do it
        res = self.hc._recover_node(ctx, node_id)

        self.assertIsNone(res)
        mock_req.assert_called_once_with(
            identity=node_id, params=self.hc.recover_action)
        self.fake_rpc.call.assert_called_once_with(ctx, 'node_recover', x_req)

    @mock.patch('senlin.objects.HealthRegistry', autospec=True)
    def test_db_create(self, mock_hrdb):
        self.hc.db_create()
        mock_hrdb.create.assert_called_once_with(
            self.hc.ctx, self.hc.cluster_id, self.hc.check_type,
            self.hc.interval, self.hc.params, self.hc.engine_id,
            self.hc.enabled)

    @mock.patch('senlin.objects.HealthRegistry', autospec=True)
    def test_db_delete(self, mock_hrdb):
        self.hc.db_delete()
        mock_hrdb.delete.assert_called_once_with(self.hc.ctx,
                                                 self.hc.cluster_id)

    @mock.patch('senlin.objects.HealthRegistry', autospec=True)
    def test_enable(self, mock_hrdb):
        self.hc.enable()
        mock_hrdb.update.assert_called_once_with(
            self.hc.ctx, self.hc.cluster_id, {'enabled': True})

    @mock.patch('senlin.objects.HealthRegistry', autospec=True)
    def test_disable(self, mock_hrdb):
        self.hc.disable()
        mock_hrdb.update.assert_called_once_with(
            self.hc.ctx, self.hc.cluster_id, {'enabled': False})


class TestRuntimeHealthRegistry(base.SenlinTestCase):

    def setUp(self):
        super(TestRuntimeHealthRegistry, self).setUp()

        mock_ctx = mock.Mock()
        self.mock_tg = mock.Mock()
        self.rhr = hm.RuntimeHealthRegistry(mock_ctx, 'ENGINE_ID',
                                            self.mock_tg)

    def create_mock_entry(self, ctx=None, engine_id='ENGINE_ID',
                          cluster_id='CID',
                          check_type=None,
                          interval=60, node_update_timeout=60, params=None,
                          enabled=True, timer=None, listener=None,
                          type=consts.POLLING):
        mock_entry = mock.Mock(
            ctx=ctx,
            engine_id=engine_id,
            cluster_id=cluster_id,
            check_type=check_type,
            interval=interval,
            node_update_timeout=node_update_timeout,
            params=params,
            enabled=enabled,
            timer=timer,
            listener=listener,
            execute_health_check=mock.Mock(),
            type=type)
        return mock_entry

    @mock.patch.object(hm, 'HealthCheck')
    def test_register_cluster(self, mock_hc):
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING])
        mock_entry.db_create = mock.Mock()
        mock_hc.return_value = mock_entry

        self.rhr.register_cluster('CID', 60, 60, {})

        self.assertEqual(mock_entry, self.rhr.registries['CID'])
        self.mock_tg.add_dynamic_timer.assert_called_once_with(
            mock_entry.execute_health_check, None, None)
        self.mock_tg.add_thread.assert_not_called()
        mock_entry.db_create.assert_called_once_with()

    @mock.patch.object(hm, 'HealthCheck')
    def test_register_cluster_failed(self, mock_hc):
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING])
        mock_entry.db_create = mock.Mock()
        mock_entry.db_delete = mock.Mock()

        mock_hc.return_value = mock_entry
        self.rhr.add_health_check = mock.Mock()
        self.rhr.add_health_check.side_effect = Exception

        self.rhr.register_cluster('CID', 60, 60, {})

        self.assertEqual(mock_entry, self.rhr.registries['CID'])
        self.mock_tg.add_dynamic_timer.assert_not_called()
        self.mock_tg.add_thread.assert_not_called()
        mock_entry.db_create.assert_called_once_with()
        mock_entry.db_delete.assert_called_once_with()

    def test_unregister_cluster_with_timer(self):
        timer = mock.Mock()
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING],
            timer=timer)
        self.rhr.registries['CID'] = mock_entry
        mock_entry.db_delete = mock.Mock()

        self.rhr.unregister_cluster('CID')

        mock_entry.db_delete.assert_called_once_with()
        timer.stop.assert_called_once_with()
        self.mock_tg.timer_done.assert_called_once_with(timer)
        self.assertIsNone(mock_entry.timer)

    def test_unregister_cluster_with_listener(self):
        listener = mock.Mock()
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING],
            listener=listener)
        self.rhr.registries['CID'] = mock_entry
        mock_entry.db_delete = mock.Mock()

        self.rhr.unregister_cluster('CID')

        mock_entry.db_delete.assert_called_once_with()
        listener.stop.assert_called_once_with()
        self.mock_tg.thread_done.assert_called_once_with(listener)
        self.assertIsNone(mock_entry.listener)

    def test_unregister_cluster_failed(self):
        listener = mock.Mock()
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING],
            listener=listener)
        self.rhr.registries['CID'] = mock_entry
        mock_entry.db_delete.side_effect = Exception

        self.rhr.unregister_cluster('CID')

        listener.stop.assert_called_once_with()
        self.mock_tg.thread_done.assert_called_once_with(listener)
        self.assertIsNone(mock_entry.listener)

    def test_enable_cluster(self):
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING],
            enabled=False)

        def mock_enable():
            mock_entry.enabled = True
            return True

        mock_entry.enable = mock_enable

        self.rhr.registries['CID'] = mock_entry

        self.rhr.enable_cluster('CID')

        self.assertTrue(mock_entry.enabled)
        self.mock_tg.add_dynamic_timer.assert_called_once_with(
            mock_entry.execute_health_check, None, None)
        self.mock_tg.add_thread.assert_not_called()

    def test_enable_cluster_failed(self):
        timer = mock.Mock()
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING],
            enabled=False, timer=timer)
        mock_entry.enable = mock.Mock()
        mock_entry.enable.side_effect = Exception

        self.rhr.registries['CID'] = mock_entry

        self.rhr.enable_cluster('CID')

        self.mock_tg.add_dynamic_timer.assert_not_called()
        self.mock_tg.add_thread.assert_not_called()
        timer.stop.assert_called_once_with()
        self.mock_tg.timer_done.assert_called_once_with(timer)

    def test_disable_cluster(self):
        timer = mock.Mock()
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING],
            enabled=True, timer=timer)

        def mock_disable():
            mock_entry.enabled = False

        mock_entry.disable = mock_disable

        self.rhr.registries['CID'] = mock_entry

        self.rhr.disable_cluster('CID')

        self.assertEqual(False, mock_entry.enabled)

        self.mock_tg.add_dynamic_timer.assert_not_called()
        self.mock_tg.add_thread.assert_not_called()
        timer.stop.assert_called_once_with()
        self.mock_tg.timer_done.assert_called_once_with(timer)

    def test_disable_cluster_failed(self):
        timer = mock.Mock()
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING],
            enabled=True, timer=timer)

        mock_entry.enable.side_effect = Exception

        self.rhr.registries['CID'] = mock_entry

        self.rhr.disable_cluster('CID')

        self.mock_tg.add_dynamic_timer.assert_not_called()
        self.mock_tg.add_thread.assert_not_called()
        timer.stop.assert_called_once_with()
        self.mock_tg.timer_done.assert_called_once_with(timer)

    def test_add_timer(self):
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING])
        self.rhr.registries['CID'] = mock_entry
        fake_timer = mock.Mock()
        self.mock_tg.add_dynamic_timer = mock.Mock()
        self.mock_tg.add_dynamic_timer.return_value = fake_timer

        self.rhr._add_timer('CID')

        self.assertEqual(fake_timer, mock_entry.timer)
        self.mock_tg.add_dynamic_timer.assert_called_once_with(
            mock_entry.execute_health_check, None, None)

    def test_add_timer_failed(self):
        fake_timer = mock.Mock()
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING], timer=fake_timer)
        self.rhr.registries['CID'] = mock_entry
        self.mock_tg.add_dynamic_timer = mock.Mock()

        self.rhr._add_timer('CID')

        self.assertEqual(fake_timer, mock_entry.timer)
        self.mock_tg.add_dynamic_timer.assert_not_called()

    @mock.patch.object(obj_profile.Profile, 'get')
    @mock.patch.object(obj_cluster.Cluster, 'get')
    def test_add_listener_nova(self, mock_cluster, mock_profile):
        cfg.CONF.set_override('nova_control_exchange', 'FAKE_NOVA_EXCHANGE',
                              group='health_manager')
        mock_entry = self.create_mock_entry(
            check_type=[consts.LIFECYCLE_EVENTS])
        self.rhr.registries['CID'] = mock_entry
        fake_listener = mock.Mock()
        x_cluster = mock.Mock(project='PROJECT_ID', profile_id='PROFILE_ID')
        mock_cluster.return_value = x_cluster
        x_profile = mock.Mock(type='os.nova.server-1.0')
        mock_profile.return_value = x_profile
        self.mock_tg.add_thread = mock.Mock()
        self.mock_tg.add_thread.return_value = fake_listener

        self.rhr._add_listener('CID')

        mock_cluster.assert_called_once_with(self.rhr.ctx, 'CID',
                                             project_safe=False)
        mock_profile.assert_called_once_with(self.rhr.ctx, 'PROFILE_ID',
                                             project_safe=False)
        self.mock_tg.add_thread.assert_called_once_with(
            hm.ListenerProc, 'FAKE_NOVA_EXCHANGE', 'PROJECT_ID', 'CID',
            mock_entry.recover_action)

    @mock.patch.object(obj_profile.Profile, 'get')
    @mock.patch.object(obj_cluster.Cluster, 'get')
    def test_add_listener_heat(self, mock_cluster, mock_profile):
        cfg.CONF.set_override('heat_control_exchange', 'FAKE_HEAT_EXCHANGE',
                              group='health_manager')
        mock_entry = self.create_mock_entry(
            check_type=[consts.LIFECYCLE_EVENTS])
        self.rhr.registries['CID'] = mock_entry
        fake_listener = mock.Mock()
        x_cluster = mock.Mock(project='PROJECT_ID', profile_id='PROFILE_ID')
        mock_cluster.return_value = x_cluster
        x_profile = mock.Mock(type='os.heat.stack-1.0')
        mock_profile.return_value = x_profile
        self.mock_tg.add_thread = mock.Mock()
        self.mock_tg.add_thread.return_value = fake_listener

        self.rhr._add_listener('CID')

        mock_cluster.assert_called_once_with(self.rhr.ctx, 'CID',
                                             project_safe=False)
        mock_profile.assert_called_once_with(self.rhr.ctx, 'PROFILE_ID',
                                             project_safe=False)
        self.mock_tg.add_thread.assert_called_once_with(
            hm.ListenerProc, 'FAKE_HEAT_EXCHANGE', 'PROJECT_ID', 'CID',
            mock_entry.recover_action)

    @mock.patch.object(obj_profile.Profile, 'get')
    @mock.patch.object(obj_cluster.Cluster, 'get')
    def test_add_listener_failed(self, mock_cluster, mock_profile):
        cfg.CONF.set_override('heat_control_exchange', 'FAKE_HEAT_EXCHANGE',
                              group='health_manager')
        fake_listener = mock.Mock()
        mock_entry = self.create_mock_entry(
            check_type=[consts.LIFECYCLE_EVENTS], listener=fake_listener)
        self.rhr.registries['CID'] = mock_entry

        x_cluster = mock.Mock(project='PROJECT_ID', profile_id='PROFILE_ID')
        mock_cluster.return_value = x_cluster
        x_profile = mock.Mock(type='os.heat.stack-1.0')
        mock_profile.return_value = x_profile
        self.mock_tg.add_thread = mock.Mock()

        self.rhr._add_listener('CID')

        mock_cluster.assert_not_called()
        mock_profile.assert_not_called()

        self.mock_tg.add_thread.assert_not_called()

    def test_add_health_check_polling(self):
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING])
        self.rhr.registries['CID'] = mock_entry
        self.rhr._add_timer = mock.Mock()
        self.rhr._add_listener = mock.Mock()

        self.rhr.add_health_check(mock_entry)

        self.rhr._add_timer.assert_called_once_with('CID')
        self.rhr._add_listener.assert_not_called()

    def test_add_health_check_events(self):
        mock_entry = self.create_mock_entry(
            check_type=[consts.LIFECYCLE_EVENTS], type=consts.EVENTS)
        self.rhr.registries['CID'] = mock_entry
        self.rhr._add_timer = mock.Mock()
        self.rhr._add_listener = mock.Mock()

        self.rhr.add_health_check(mock_entry)

        self.rhr._add_timer.assert_not_called()
        self.rhr._add_listener.assert_called_once_with('CID')

    def test_add_health_check_disabled(self):
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING], enabled=False)
        self.rhr.registries['CID'] = mock_entry
        self.rhr._add_timer = mock.Mock()
        self.rhr._add_listener = mock.Mock()

        self.rhr.add_health_check(mock_entry)

        self.rhr._add_timer.assert_not_called()
        self.rhr._add_listener.assert_not_called()

    def test_add_health_check_timer_exists(self):
        fake_timer = mock.Mock()
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING], timer=fake_timer)
        self.rhr.registries['CID'] = mock_entry
        self.rhr._add_timer = mock.Mock()
        self.rhr._add_listener = mock.Mock()

        self.rhr.add_health_check(mock_entry)

        self.rhr._add_timer.assert_not_called()
        self.rhr._add_listener.assert_not_called()

    def test_remove_health_check_timer(self):
        fake_timer = mock.Mock()
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING], timer=fake_timer)
        self.rhr.registries['CID'] = mock_entry

        self.rhr.remove_health_check(mock_entry)

        fake_timer.stop.assert_called_once_with()
        self.mock_tg.timer_done.assert_called_once_with(fake_timer)
        self.mock_tg.thread_done.assert_not_called()
        self.assertIsNone(mock_entry.timer)

    def test_remove_health_check_listener(self):
        fake_listener = mock.Mock()
        mock_entry = self.create_mock_entry(
            check_type=[consts.NODE_STATUS_POLLING], listener=fake_listener)
        self.rhr.registries['CID'] = mock_entry

        self.rhr.remove_health_check(mock_entry)

        fake_listener.stop.assert_called_once_with()
        self.mock_tg.timer_done.assert_not_called()
        self.mock_tg.thread_done.assert_called_once_with(fake_listener)
        self.assertIsNone(mock_entry.listener)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_context import context as oslo_context
from oslo_log import log as logging
import oslo_messaging

from senlin.common import consts
from senlin.common import messaging

LOG = logging.getLogger(__name__)

OPERATIONS = (
    START_ACTION, CANCEL_ACTION, STOP
) = (
    'start_action', 'cancel_action', 'stop'
)


def notify(method, engine_id=None, **kwargs):
    """Send notification to dispatcher.

    Note that dispatcher is an engine internal communication. We are not using
    versioned object serialization at this level.

    :param method: remote method to call
    :param engine_id: dispatcher to notify; None implies broadcast
    """
    client = messaging.get_rpc_client(consts.ENGINE_TOPIC, cfg.CONF.host)

    if engine_id:
        # Notify specific dispatcher identified by engine_id
        call_context = client.prepare(server=engine_id)
    else:
        # Broadcast to all disptachers
        call_context = client.prepare(fanout=True)

    try:
        # We don't use ctext parameter in action progress
        # actually. But since RPCClient.call needs this param,
        # we use oslo current context here.
        call_context.cast(oslo_context.get_current(), method, **kwargs)
        return True
    except oslo_messaging.MessagingTimeout:
        return False


def start_action(engine_id=None, **kwargs):
    return notify(START_ACTION, engine_id, **kwargs)



#!/usr/bin/env python
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Senlin Conductor.
"""
import sys

from oslo_log import log as logging
from oslo_reports import guru_meditation_report as gmr
from oslo_service import service

from senlin.common import config
from senlin.common import consts
from senlin.common import messaging
from senlin.common import profiler
import senlin.conf
from senlin import objects
from senlin import version

CONF = senlin.conf.CONF


def main():
    config.parse_args(sys.argv, 'senlin-conductor')
    logging.setup(CONF, 'senlin-conductor')
    logging.set_defaults()
    gmr.TextGuruMeditation.setup_autorun(version)
    objects.register_all()
    messaging.setup()

    from senlin.conductor import service as conductor

    profiler.setup('senlin-conductor', CONF.host)
    srv = conductor.ConductorService(CONF.host, consts.CONDUCTOR_TOPIC)
    launcher = service.launch(CONF, srv,
                              workers=CONF.conductor.workers,
                              restart_method='mutate')
    # the following periodic tasks are intended serve as HA checking
    # srv.create_periodic_tasks()
    launcher.wait()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Health Manager.

Health Manager is responsible for monitoring the health of the clusters and
trigger corresponding actions to recover the clusters based on the pre-defined
health policies.
"""
from collections import defaultdict
from collections import namedtuple
import eventlet
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from oslo_utils import timeutils
import re
import tenacity

from senlin.common import consts
from senlin.common import context
from senlin.common import messaging as rpc
from senlin.common import utils
from senlin.engine import node as node_mod
from senlin.engine.notifications import heat_endpoint
from senlin.engine.notifications import nova_endpoint
from senlin import objects
from senlin.rpc import client as rpc_client

LOG = logging.getLogger(__name__)


def chase_up(start_time, interval, name='Poller'):
    """Utility function to check if there are missed intervals.

    :param start_time: A time object representing the starting time.
    :param interval: An integer specifying the time interval in seconds.
    :param name: Name of the caller for identification in logs.
    :returns: Number of seconds to sleep before next round.
    """
    end_time = timeutils.utcnow(True)
    elapsed = timeutils.delta_seconds(start_time, end_time)
    # check if we have missed any intervals?
    missed = int((elapsed - 0.0000001) / interval)
    if missed >= 1:
        LOG.warning("%s missed %s intervals for checking", name, missed)
    return (missed + 1) * interval - elapsed


def ListenerProc(exchange, project_id, cluster_id, recover_action):
    """Thread procedure for running an event listener.

    :param exchange: The control exchange for a target service.
    :param project_id: The ID of the project to filter.
    :param cluster_id: The ID of the cluster to filter.
    :param recover_action: The health policy action name.
    """
    transport = messaging.get_notification_transport(cfg.CONF)

    if exchange == cfg.CONF.health_manager.nova_control_exchange:
        endpoint = nova_endpoint.NovaNotificationEndpoint(
            project_id, cluster_id, recover_action
        )

    else:
        endpoint = heat_endpoint.HeatNotificationEndpoint(
            project_id, cluster_id, recover_action
        )

    listener = messaging.get_notification_listener(
        transport, [endpoint.target], [endpoint], executor='threading',
        pool='senlin-listeners'
    )

    listener.start()


class HealthCheckType(object):
    @staticmethod
    def factory(detection_type, cid, interval, params):
        node_update_timeout = params['node_update_timeout']
        detection_params = [
            p for p in params['detection_modes']
            if p['type'] == detection_type
        ]
        if len(detection_params) != 1:
            raise Exception(
                'The same detection mode cannot be used more than once in the '
                'same policy. Encountered {} instances of '
                'type {}.'.format(len(detection_params), detection_type)
            )

        if detection_type == consts.NODE_STATUS_POLLING:
            return NodePollStatusHealthCheck(
                cid, interval, node_update_timeout, detection_params[0])
        elif detection_type == consts.NODE_STATUS_POLL_URL:
            return NodePollUrlHealthCheck(
                cid, interval, node_update_timeout, detection_params[0])
        elif detection_type == consts.HYPERVISOR_STATUS_POLLING:
            return HypervisorPollStatusHealthCheck(
                cid, interval, node_update_timeout, detection_params[0])
        else:
            raise Exception(
                'Invalid detection type: {}'.format(detection_type))

    def __init__(self, cluster_id, interval, node_update_timeout, params):
        """Initialize HealthCheckType

        :param ctx:
        :param cluster_id: The UUID of the cluster to be checked.
        :param params: Parameters specific to poll url or recovery action.
        """
        self.cluster_id = cluster_id
        self.interval = interval
        self.node_update_timeout = node_update_timeout
        self.params = params

    def run_health_check(self, ctx, node):
        """Run health check on node

        :returns: True if node is healthy. False otherwise.
        """
        pass

    def _node_within_grace_period(self, node):
        """Check if current time is within the node_update_timeout grace period

        :returns: True if current time is less than node_update_timeout since
            last node update action. False otherwise.
        """

        node_last_updated = node.updated_at or node.init_at
        if timeutils.is_older_than(node_last_updated,
                                   self.node_update_timeout):
            # node was last updated more than node_update_timeout seconds ago
            # -> we are outside the grace period
            LOG.info("%s was updated at %s which is more "
                     "than %d secs ago. Mark node as unhealthy.",
                     node.name, node_last_updated,
                     self.node_update_timeout)
            return False
        else:
            # node was last updated less than node_update_timeout seconds ago
            # -> we are inside the grace period
            LOG.info("%s was updated at %s which is less "
                     "than %d secs ago. Mark node as healthy.",
                     node.name, node_last_updated,
                     self.node_update_timeout)
            return True


class NodePollStatusHealthCheck(HealthCheckType):
    def run_health_check(self, ctx, node):
        """Routine to be executed for polling node status.

        :returns: True if node is healthy. False otherwise.
        """
        try:
            # create engine node from db node
            entity = node_mod.Node._from_object(ctx, node)

            # If health check returns True, return True to mark node as
            # healthy. Else return True to mark node as healthy if we are still
            # within the node's grace period to allow the node to warm-up.
            # Return False to mark the node as unhealthy if we are outside the
            # grace period.

            return (entity.do_healthcheck(ctx, consts.NODE_STATUS_POLLING) or
                    self._node_within_grace_period(node))
        except Exception as ex:
            LOG.warning(
                'Error when performing health check on node %s: %s',
                node.id, ex
            )

            # treat node as healthy when an exception is encountered
            return True


class HypervisorPollStatusHealthCheck(HealthCheckType):
    def run_health_check(self, ctx, node):
        """Routine to be executed for polling hypervisor status.

        :returns: True if node is healthy. False otherwise.
        """
        try:
            # create engine node from db node
            entity = node_mod.Node._from_object(ctx, node)

            # If health check returns True, return True to mark node as
            # healthy. Else return True to mark node as healthy if we are still
            # within the node's grace period to allow the node to warm-up.
            # Return False to mark the node as unhealthy if we are outside the
            # grace period.

            return (entity.do_healthcheck(ctx,
                                          consts.HYPERVISOR_STATUS_POLLING) or
                    self._node_within_grace_period(node))
        except Exception as ex:
            LOG.warning(
                'Error when performing health check on node %s: %s',
                node.id, ex
            )

            # treat node as healthy when an exception is encountered
            return True


class NodePollUrlHealthCheck(HealthCheckType):
    @staticmethod
    def convert_detection_tuple(dictionary):
        return namedtuple('DetectionMode', dictionary.keys())(**dictionary)

    def _expand_url_template(self, url_template, node):
        """Expands parameters in an URL template

        :param url_template:
            A string containing parameters that will be expanded. Currently
            only the {nodename} parameter is supported, which will be replaced
            by the actual node name.
        :param node: The DB object for the node to use for parameter expansion
        :returns: A string containing the expanded URL
        """

        nodename_pattern = re.compile("(\{nodename\})")
        url = nodename_pattern.sub(node.name, url_template)

        return url

    def _poll_url(self, url, node):
        verify_ssl = self.params['poll_url_ssl_verify']
        conn_error_as_unhealthy = self.params[
            'poll_url_conn_error_as_unhealthy']
        expected_resp_str = self.params['poll_url_healthy_response']
        retry_interval = self.params['poll_url_retry_interval']

        timeout = max(retry_interval * 0.1, 1)

        try:
            result = utils.url_fetch(url, timeout=timeout,
                                     verify=verify_ssl)
        except Exception as ex:
            if conn_error_as_unhealthy:
                LOG.info("%s for %s: connection error when polling URL (%s)",
                         consts.POLL_URL_FAIL, node.name, ex)
                return False
            else:
                LOG.info("%s for %s: ignoring connection error when polling "
                         "URL (%s)",
                         consts.POLL_URL_PASS, node.name, ex)
                return True

        if not re.search(expected_resp_str, result):
            LOG.info("%s for %s: did not find expected response string %s in "
                     "URL result (%s)",
                     consts.POLL_URL_FAIL, node.name, expected_resp_str,
                     result)
            return False

        LOG.info("%s for %s: matched expected response string.",
                 consts.POLL_URL_PASS, node.name)
        return True

    def run_health_check(self, ctx, node):
        """Routine to check a node status from a url and recovery if necessary

        :param node: The node to be checked.
        :returns: True if node is healthy. False otherwise.
        """

        max_unhealthy_retry = self.params['poll_url_retry_limit']
        retry_interval = self.params['poll_url_retry_interval']

        def _return_last_value(retry_state):
            return retry_state.outcome.result()

        @tenacity.retry(
            retry=tenacity.retry_if_result(lambda x: x is False),
            wait=tenacity.wait_fixed(retry_interval),
            retry_error_callback=_return_last_value,
            stop=tenacity.stop_after_attempt(max_unhealthy_retry)
        )
        def _poll_url_with_retry(url):
            return self._poll_url(url, node)

        try:
            if node.status != consts.NS_ACTIVE:
                LOG.info("%s for %s: node is not in ACTIVE state, so skip "
                         "poll url",
                         consts.POLL_URL_PASS, node.name)
                return True

            url_template = self.params['poll_url']
            url = self._expand_url_template(url_template, node)

            # If health check returns True, return True to mark node as
            # healthy. Else return True to mark node as healthy if we are still
            # within the node's grace period to allow the node to warm-up.
            # Return False to mark the node as unhealthy if we are outside the
            # grace period.

            return (_poll_url_with_retry(url) or
                    self._node_within_grace_period(node))
        except Exception as ex:
            LOG.warning(
                "%s for %s: Ignoring error on poll URL: %s",
                consts.POLL_URL_PASS, node.name, ex
            )

            # treat node as healthy when an exception is encountered
            return True


class HealthCheck(object):

    def __init__(self, ctx, engine_id, cluster_id, check_type, interval,
                 node_update_timeout, params, enabled):
        self.rpc_client = rpc_client.get_engine_client()
        self.ctx = ctx
        self.engine_id = engine_id

        self.cluster_id = cluster_id
        self.check_type = check_type
        self.interval = interval
        self.node_update_timeout = node_update_timeout
        self.params = params
        self.enabled = enabled
        self.timer = None
        self.listener = None

        self.health_check_types = []
        self.recover_action = {}
        self.type = None
        self.get_health_check_types()
        self.get_recover_actions()

    def get_health_check_types(self):
        polling_types = [consts.NODE_STATUS_POLLING,
                         consts.NODE_STATUS_POLL_URL,
                         consts.HYPERVISOR_STATUS_POLLING]

        detection_types = self.check_type.split(',')
        if all(check in polling_types for check in detection_types):
            interval = min(self.interval, cfg.CONF.check_interval_max)
            for check in detection_types:
                self.health_check_types.append(
                    HealthCheckType.factory(
                        check, self.cluster_id, interval, self.params
                    )
                )
            self.type = consts.POLLING
        elif (len(detection_types) == 1 and
              detection_types[0] == consts.LIFECYCLE_EVENTS):
            self.type = consts.EVENTS

    def get_recover_actions(self):
        if 'node_delete_timeout' in self.params:
            self.recover_action['delete_timeout'] = self.params[
                'node_delete_timeout']
        if 'node_force_recreate' in self.params:
            self.recover_action['force_recreate'] = self.params[
                'node_force_recreate']
        if 'recover_action' in self.params:
            rac = self.params['recover_action']
            for operation in rac:
                self.recover_action['operation'] = operation.get('name')

    def execute_health_check(self):
        start_time = timeutils.utcnow(True)

        try:
            if not self.health_check_types:
                LOG.error("No health check types found for cluster: %s",
                          self.cluster_id)
                return chase_up(start_time, self.interval)

            cluster = objects.Cluster.get(self.ctx, self.cluster_id,
                                          project_safe=False)
            if not cluster:
                LOG.warning("Cluster (%s) is not found.", self.cluster_id)
                return chase_up(start_time, self.interval)

            ctx = context.get_service_context(user_id=cluster.user,
                                              project_id=cluster.project)

            actions = []

            # loop through nodes and run all health checks on each node
            nodes = objects.Node.get_all_by_cluster(ctx, self.cluster_id)

            for node in nodes:
                action = self._check_node_health(ctx, node, cluster)
                if action:
                    actions.append(action)

            for a in actions:
                # wait for action to complete
                res, reason = self._wait_for_action(
                    ctx, a['action'], self.node_update_timeout)
                if not res:
                    LOG.warning("Node recovery action %s did not complete "
                                "within specified timeout: %s", a['action'],
                                reason)

            if len(actions) == 0:
                LOG.info("Health check passed for all nodes in cluster %s.",
                         self.cluster_id)
        except Exception as ex:
            LOG.warning("Error while performing health check: %s", ex)

        finally:
            return chase_up(start_time, self.interval)

    def _check_node_health(self, ctx, node, cluster):
        node_is_healthy = True

        if self.params['recovery_conditional'] == consts.ANY_FAILED:
            # recovery happens if any detection mode fails
            # i.e. the inverse logic is that node is considered healthy
            # if all detection modes pass
            node_is_healthy = all(
                hc.run_health_check(ctx, node)
                for hc in self.health_check_types)
        elif self.params['recovery_conditional'] == consts.ALL_FAILED:
            # recovery happens if all detection modes fail
            # i.e. the inverse logic is that node is considered healthy
            # if any detection mode passes
            node_is_healthy = any(
                hc.run_health_check(ctx, node)
                for hc in self.health_check_types)
        else:
            raise Exception("%s is an invalid recovery conditional" %
                            self.params['recovery_conditional'])

        if not node_is_healthy:
            LOG.info("Health check failed for %s in %s and "
                     "recovery has started.",
                     node.name, cluster.name)
            return self._recover_node(ctx, node.id)

    def _wait_for_action(self, ctx, action_id, timeout):
        req = objects.ActionGetRequest(identity=action_id)
        action = {}
        with timeutils.StopWatch(timeout) as timeout_watch:
            while not timeout_watch.expired():
                action = self.rpc_client.call(ctx, 'action_get', req)
                if action['status'] in [consts.ACTION_SUCCEEDED,
                                        consts.ACTION_FAILED,
                                        consts.ACTION_CANCELLED]:
                    break
                eventlet.sleep(2)

        if not action:
            return False, "Failed to retrieve action."

        elif action['status'] == consts.ACTION_SUCCEEDED:
            return True, ""

        elif (action['status'] == consts.ACTION_FAILED or
              action['status'] == consts.ACTION_CANCELLED):
            return False, "Cluster check action failed or cancelled"

        return False, ("Timeout while waiting for node recovery action to "
                       "finish")

    def _recover_node(self, ctx, node_id):
        """Recover node

        :returns: Recover action
        """
        try:
            req = objects.NodeRecoverRequest(identity=node_id,
                                             params=self.recover_action)

            return self.rpc_client.call(ctx, 'node_recover', req)
        except Exception as ex:
            LOG.error("Error when performing node recovery for %s: %s",
                      node_id, ex)
            return None

    def db_create(self):
        try:
            objects.HealthRegistry.create(
                self.ctx, self.cluster_id, self.check_type, self.interval,
                self.params, self.engine_id, self.enabled)
            return True
        except Exception as ex:
            LOG.error("Error while adding health entry for cluster %s to "
                      "database: %s", self.cluster_id, ex)
            return False

    def db_delete(self):
        try:
            objects.HealthRegistry.delete(self.ctx, self.cluster_id)
            return True
        except Exception as ex:
            LOG.error("Error while removing health entry for cluster %s from "
                      "database: %s", self.cluster_id, ex)
            return False

    def enable(self):
        try:
            objects.HealthRegistry.update(self.ctx, self.cluster_id,
                                          {'enabled': True})
            self.enabled = True
            return True
        except Exception as ex:
            LOG.error("Error while enabling health entry for cluster %s: %s",
                      self.cluster_id, ex)
            return False

    def disable(self):
        try:
            objects.HealthRegistry.update(self.ctx, self.cluster_id,
                                          {'enabled': False})
            self.enabled = False
            return True
        except Exception as ex:
            LOG.error("Error while disabling health entry for cluster %s: %s",
                      self.cluster_id, ex)
            return False


class RuntimeHealthRegistry(object):
    def __init__(self, ctx, engine_id, thread_group):
        self.ctx = ctx
        self.engine_id = engine_id
        self.rt = {}
        self.tg = thread_group
        self.health_check_types = defaultdict(lambda: [])

    @property
    def registries(self):
        return self.rt

    def register_cluster(self, cluster_id, interval=None,
                         node_update_timeout=None, params=None,
                         enabled=True):
        """Register cluster to health registry.

        :param cluster_id: The ID of the cluster to be registered.
        :param interval: An optional integer indicating the length of checking
                         periods in seconds.
        :param node_update_timeout: Timeout to wait for node action to
                                    complete.
        :param dict params: Other parameters for the health check.
        :param enabled: Boolean indicating if the health check is enabled.
        :return: RuntimeHealthRegistry object for cluster
        """
        params = params or {}

        # extract check_type from params
        check_type = ""
        if 'detection_modes' in params:
            check_type = ','.join([
                NodePollUrlHealthCheck.convert_detection_tuple(d).type
                for d in params['detection_modes']
            ])

        # add node_update_timeout to params
        params['node_update_timeout'] = node_update_timeout
        entry = None
        try:
            entry = HealthCheck(
                ctx=self.ctx,
                engine_id=self.engine_id,
                cluster_id=cluster_id,
                check_type=check_type,
                interval=interval,
                node_update_timeout=node_update_timeout,
                params=params,
                enabled=enabled
            )
            if entry.db_create():
                self.registries[cluster_id] = entry
                self.add_health_check(self.registries[cluster_id])
        except Exception as ex:
            LOG.error("Error while trying to register cluster for health "
                      "check %s: %s", cluster_id, ex)
            if entry:
                entry.db_delete()

    def unregister_cluster(self, cluster_id):
        """Unregister a cluster from health registry.

        :param cluster_id: The ID of the cluster to be unregistered.
        :return: RuntimeHealthRegistry object for the cluster being
                 unregistered.
        """
        entry = None
        try:
            if cluster_id in self.registries:
                entry = self.registries.pop(cluster_id)
                entry.db_delete()
        except Exception as ex:
            LOG.error("Error while trying to unregister cluster from health "
                      "check %s: %s", cluster_id, ex)
        finally:
            if entry:
                self.remove_health_check(entry)

    def enable_cluster(self, cluster_id):
        """Update the status of a cluster to enabled in the health registry.

        :param cluster_id: The ID of the cluster to be enabled.
        """
        LOG.info("Enabling health check for cluster %s.", cluster_id)
        try:
            if cluster_id in self.registries:
                if self.registries[cluster_id].enable():
                    self.add_health_check(self.registries[cluster_id])
            else:
                LOG.error("Unable to enable cluster for health checking: %s",
                          cluster_id)
        except Exception as ex:
            LOG.error("Error while enabling health check for cluster %s: %s",
                      cluster_id, ex)
            if cluster_id in self.registries:
                self.remove_health_check(self.registries[cluster_id])

    def disable_cluster(self, cluster_id):
        """Update the status of a cluster to disabled in the health registry.

        :param cluster_id: The ID of the cluster to be disabled.
        :return: None.
        """
        LOG.info("Disabling health check for cluster %s.", cluster_id)
        try:
            if cluster_id in self.registries:
                self.registries[cluster_id].disable()
            else:
                LOG.error("Unable to disable cluster for health checking: %s",
                          cluster_id)
        except Exception as ex:
            LOG.error("Error while disabling health check for cluster %s: %s",
                      cluster_id, ex)
        finally:
            if cluster_id in self.registries:
                self.remove_health_check(self.registries[cluster_id])

    def _add_timer(self, cluster_id):
        entry = self.registries[cluster_id]
        if entry.timer:
            LOG.error("Health check for cluster %s already exists", cluster_id)
            return None
        timer = self.tg.add_dynamic_timer(entry.execute_health_check, None,
                                          None)
        if timer:
            entry.timer = timer
        else:
            LOG.error("Error creating timer for cluster: %s", cluster_id)

    def _add_listener(self, cluster_id):
        entry = self.registries[cluster_id]
        if entry.listener:
            LOG.error("Listener for cluster %s already exists", cluster_id)
            return

        cluster = objects.Cluster.get(self.ctx, cluster_id, project_safe=False)
        if not cluster:
            LOG.warning("Cluster (%s) is not found.", cluster_id)
            return
        profile = objects.Profile.get(self.ctx, cluster.profile_id,
                                      project_safe=False)
        profile_type = profile.type.split('-')[0]
        if profile_type == 'os.nova.server':
            exchange = cfg.CONF.health_manager.nova_control_exchange
        elif profile_type == 'os.heat.stack':
            exchange = cfg.CONF.health_manager.heat_control_exchange
        else:
            return

        project = cluster.project
        listener = self.tg.add_thread(ListenerProc, exchange, project,
                                      cluster_id, entry.recover_action)
        if listener:
            entry.listener = listener
        else:
            LOG.error("Error creating listener for cluster: %s", cluster_id)

    def add_health_check(self, entry):
        """Add a health check to the RuntimeHealthRegistry.

        This method creates a timer/thread based on the type of health check
        being added.

        :param entry: Entry to add to the registry.
        :return: None
        """
        if entry.cluster_id in self.registries:
            if not entry.enabled:
                return
            elif entry.timer:
                LOG.error("Health check for cluster %s already exists",
                          entry.cluster_id)
                return
        else:
            LOG.error("Unable to add health check for cluster: %s",
                      entry.cluster_id)
            return

        if entry.type == consts.POLLING:
            self._add_timer(entry.cluster_id)
        elif entry.type == consts.EVENTS:
            LOG.info("Start listening events for cluster (%s).",
                     entry.cluster_id)
            self._add_listener(entry.cluster_id)
        else:
            LOG.error("Cluster %(id)s type %(type)s is invalid.",
                      {'id': entry.cluster_id, 'type': entry.type})

    def remove_health_check(self, entry):
        """Remove a health check for the RuntimeHealthRegistry.

        This method stops and removes the timer/thread based to the type of
        health check being removed.

        :param entry:
        :return: None
        """
        if entry.timer:
            # stop timer
            entry.timer.stop()

            try:
                # tell threadgroup to remove timer
                self.tg.timer_done(entry.timer)
            except ValueError:
                pass
            finally:
                entry.timer = None

        if entry.listener:
            try:
                self.tg.thread_done(entry.listener)
                entry.listener.stop()
            except ValueError:
                pass
            finally:
                entry.listener = None

    def load_runtime_registry(self):
        """Load the initial runtime registry with a DB scan."""
        db_registries = objects.HealthRegistry.claim(self.ctx, self.engine_id)

        for registry in db_registries:
            if registry.cluster_id in self.registries:
                LOG.warning("Skipping duplicate health check for cluster: %s",
                            registry.cluster_id)
            # Claiming indicates we claim a health registry who's engine was
            # dead, and we will update the health registry's engine_id with
            # current engine id. But we may not start check always.
            entry = HealthCheck(
                ctx=self.ctx,
                engine_id=self.engine_id,
                cluster_id=registry.cluster_id,
                check_type=registry.check_type,
                interval=registry.interval,
                node_update_timeout=registry.params['node_update_timeout'],
                params=registry.params,
                enabled=registry.enabled
            )

            LOG.info("Loading cluster %(c)s enabled=%(e)s for "
                     "health monitoring",
                     {'c': registry.cluster_id, 'e': registry.enabled})
            self.registries[registry.cluster_id] = entry
            if registry.enabled:
                self.add_health_check(self.registries[registry.cluster_id])


def notify(engine_id, method, **kwargs):
    """Send notification to health manager service.

    Note that the health manager only handles JSON type of parameter passing.

    :param engine_id: dispatcher to notify; broadcast if value is None
    :param method: remote method to call
    """
    timeout = cfg.CONF.engine_life_check_timeout
    client = rpc.get_rpc_client(consts.HEALTH_MANAGER_TOPIC, None)

    if engine_id:
        # Notify specific dispatcher identified by engine_id
        call_context = client.prepare(timeout=timeout, server=engine_id)
    else:
        # Broadcast to all disptachers
        call_context = client.prepare(timeout=timeout)

    ctx = context.get_admin_context()

    try:
        call_context.call(ctx, method, **kwargs)
        return True
    except messaging.MessagingTimeout:
        return False


def register(cluster_id, engine_id=None, **kwargs):
    params = kwargs.pop('params', {})
    interval = kwargs.pop('interval', cfg.CONF.periodic_interval)
    node_update_timeout = kwargs.pop('node_update_timeout', 300)
    enabled = kwargs.pop('enabled', True)
    return notify(engine_id, 'register_cluster',
                  cluster_id=cluster_id,
                  interval=interval,
                  node_update_timeout=node_update_timeout,
                  params=params,
                  enabled=enabled)


def unregister(cluster_id):
    engine_id = get_manager_engine(cluster_id)
    if engine_id:
        return notify(engine_id, 'unregister_cluster', cluster_id=cluster_id)
    return True


def enable(cluster_id, **kwargs):
    engine_id = get_manager_engine(cluster_id)
    if engine_id:
        return notify(engine_id, 'enable_cluster', cluster_id=cluster_id,
                      params=kwargs)
    return False


def disable(cluster_id, **kwargs):
    engine_id = get_manager_engine(cluster_id)
    if engine_id:
        return notify(engine_id, 'disable_cluster', cluster_id=cluster_id,
                      params=kwargs)
    return False


def get_manager_engine(cluster_id):
    ctx = context.get_admin_context()

    registry = objects.HealthRegistry.get(ctx, cluster_id)
    if not registry:
        return None

    return registry.engine_id



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log as logging
from oslo_service import service

from oslo_service import sslutils
from oslo_service import wsgi
from oslo_utils import netutils
from oslo_utils import uuidutils

from senlin.common import context as senlin_context
import senlin.conf
from senlin.objects import service as service_obj
from senlin import version

CONF = senlin.conf.CONF
LOG = logging.getLogger(__name__)


class Service(service.Service):
    def __init__(self, name, host, topic, threads=None):
        self.tg = None
        super(Service, self).__init__(threads or 1000)
        self.name = name
        self.host = host
        self.topic = topic

        self.server = None
        self.service_id = None
        self.cleanup_timer = None
        self.cleanup_count = 0
        self.service_report_timer = None

        # Start the service cleanup process. This is only going to be
        # running on the main process.
        if self.tg:
            self.cleanup_timer = self.tg.add_timer(
                CONF.periodic_interval, self.service_manage_cleanup
            )

    def start(self):
        super(Service, self).start()
        self.service_id = uuidutils.generate_uuid()
        LOG.info(
            'Starting %(name)s service (version: %(version)s '
            'id: %(service_id)s)',
            {
                'name': self.name,
                'version': version.version_info.version_string(),
                'service_id': self.service_id,
            }
        )
        ctx = senlin_context.get_admin_context()
        service_obj.Service.create(
            ctx, self.service_id, self.host, self.name, self.topic
        )
        self.service_report_timer = self.tg.add_timer(
            CONF.periodic_interval, self.service_manage_report
        )

    def stop(self, graceful=True):
        LOG.info(
            'Stopping %(name)s service (id: %(service_id)s)',
            {
                'name': self.name,
                'service_id': self.service_id or 'main',
            }
        )
        if self.service_report_timer:
            self.service_report_timer.stop()
            self.service_report_timer = None
        if self.cleanup_timer:
            self.cleanup_timer.stop()
            self.cleanup_timer = None
        if self.service_id:
            service_obj.Service.delete(self.service_id)
        super(Service, self).stop(graceful)

    def service_manage_cleanup(self):
        self.cleanup_count += 1
        try:
            ctx = senlin_context.get_admin_context()
            services = service_obj.Service.get_all_expired(
                ctx, self.name
            )
            for svc in services:
                LOG.info(
                    'Breaking locks for dead service %(name)s '
                    '(id: %(service_id)s)',
                    {
                        'name': self.name,
                        'service_id': svc['id'],
                    }
                )
                service_obj.Service.gc_by_engine(svc['id'])
                LOG.info(
                    'Done breaking locks for service %(name)s '
                    '(id: %(service_id)s)',
                    {
                        'name': self.name,
                        'service_id': svc['id'],
                    }
                )
                service_obj.Service.delete(svc['id'])
        except Exception as ex:
            LOG.error(
                'Error while cleaning up service %(name)s: %(ex)s',
                {
                    'name': self.name,
                    'ex': ex,
                }
            )

        # The clean-up process runs during service startup and will over
        # multiple attempts check to see if any services have reach the
        # deadline and if so remove them. This is only done on startup, or
        # after a service recovers from a crash.
        if self.cleanup_count >= 5:
            self.cleanup_timer.stop()
            self.cleanup_timer = None
            LOG.info('Finished cleaning up dead services.')
        else:
            LOG.info('Service clean-up attempt count: %s', self.cleanup_count)

    def service_manage_report(self):
        try:
            ctx = senlin_context.get_admin_context()
            service_obj.Service.update(ctx, self.service_id)
        except Exception as ex:
            LOG.error(
                'Error while updating service %(name)s: %(ex)s',
                {
                    'name': self.name,
                    'ex': ex,
                }
            )


class WSGIService(service.Service):
    def __init__(self, app, name, listen, max_url_len=None):
        super(WSGIService, self).__init__(CONF.senlin_api.threads)
        self.app = app
        self.name = name

        self.listen = listen

        self.servers = []

        for address in self.listen:
            host, port = netutils.parse_host_port(address)
            server = wsgi.Server(
                CONF, name, app,
                host=host,
                port=port,
                pool_size=CONF.senlin_api.threads,
                use_ssl=sslutils.is_enabled(CONF),
                max_url_len=max_url_len
            )

            self.servers.append(server)

    def start(self):
        for server in self.servers:
            server.start()
        super(WSGIService, self).start()

    def stop(self, graceful=True):
        for server in self.servers:
            server.stop()
        super(WSGIService, self).stop(graceful)

    def wait(self):
        for server in self.servers:
            server.wait()
        super(WSGIService, self).wait()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_utils import timeutils
from osprofiler import profiler

from senlin.common import consts
from senlin.common import context
from senlin.common import messaging as rpc
from senlin.common import service
from senlin.engine import health_manager

LOG = logging.getLogger(__name__)
CONF = cfg.CONF


@profiler.trace_cls("rpc")
class HealthManagerService(service.Service):
    def __init__(self, host, topic):
        super(HealthManagerService, self).__init__(
            self.service_name, host, topic,
            threads=CONF.health_manager.threads
        )
        self.version = consts.RPC_API_VERSION

        self.ctx = context.get_admin_context()

        # The following are initialized here and will be assigned in start()
        # which happens after the fork when spawning multiple worker processes
        self.health_registry = None
        self.target = None

    @property
    def service_name(self):
        return 'senlin-health-manager'

    def start(self):
        super(HealthManagerService, self).start()

        self.health_registry = health_manager.RuntimeHealthRegistry(
            ctx=self.ctx, engine_id=self.service_id,
            thread_group=self.tg
        )
        self.target = oslo_messaging.Target(server=self.service_id,
                                            topic=self.topic,
                                            version=self.version)
        self.server = rpc.get_rpc_server(self.target, self)
        self.server.start()

        self.tg.add_dynamic_timer(self.task, None, cfg.CONF.periodic_interval)

    def stop(self, graceful=False):
        if self.server:
            self.server.stop()
            self.server.wait()
        super(HealthManagerService, self).stop(graceful)

    def task(self):
        """Task that is queued on the health manager thread group.

        The task is here so that the service always has something to wait()
        on, or else the process will exit.
        """
        start_time = timeutils.utcnow(True)

        try:
            self.health_registry.load_runtime_registry()
        except Exception as ex:
            LOG.error("Failed when loading runtime for health manager: %s", ex)
        return health_manager.chase_up(
            start_time, cfg.CONF.periodic_interval, name='Health manager task'
        )

    def listening(self, ctx):
        """Respond to confirm that the rpc service is still alive."""
        return True

    def register_cluster(self, ctx, cluster_id, interval=None,
                         node_update_timeout=None, params=None,
                         enabled=True):
        """Register a cluster for health checking.

        :param ctx: The context of notify request.
        :param cluster_id: The ID of the cluster to be unregistered.
        :param interval: Interval of the health check.
        :param node_update_timeout: Time to wait before declairing a node
        unhealthy.
        :param params: Params to be passed to health check.
        :param enabled: Set's if the health check is enabled or disabled.
        :return: None
        """
        LOG.info("Registering health check for cluster %s.", cluster_id)
        self.health_registry.register_cluster(
            cluster_id=cluster_id,
            interval=interval,
            node_update_timeout=node_update_timeout,
            params=params,
            enabled=enabled)

    def unregister_cluster(self, ctx, cluster_id):
        """Unregister a cluster from health checking.

        :param ctx: The context of notify request.
        :param cluster_id: The ID of the cluster to be unregistered.
        :return: None
        """
        LOG.info("Unregistering health check for cluster %s.", cluster_id)
        self.health_registry.unregister_cluster(cluster_id)

    def enable_cluster(self, ctx, cluster_id, params=None):
        self.health_registry.enable_cluster(cluster_id)

    def disable_cluster(self, ctx, cluster_id, params=None):
        self.health_registry.disable_cluster(cluster_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
# -*- coding: utf-8 -*-

from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives
from functools import cmp_to_key
from oslo_utils import importutils
from sphinx.util import logging

from senlin.common import schema

LOG = logging.getLogger(__name__)


class SchemaDirective(rst.Directive):
    required_arguments = 0
    optional_arguments = 0
    final_argument_whitespace = True
    option_spec = {'package': directives.unchanged}
    has_content = False
    add_index = True
    section_title = 'Spec'
    properties_only = False

    def run(self):
        """Build doctree nodes consisting for the specified schema class

        :returns: doctree node list
        """

        # gives you access to the options of the directive
        options = self.options

        content = []

        # read in package class
        obj = importutils.import_class(options['package'])

        # skip other spec properties if properties_only is True
        if not self.properties_only:
            section = self._create_section(content, 'spec',
                                           title=self.section_title)

            # create version section
            version_section = self._create_section(section, 'version',
                                                   title='Latest Version')
            field = nodes.line('', obj.VERSION)
            version_section.append(field)

            # build versions table
            version_tbody = self._build_table(
                section, 'Available Versions',
                ['Version', 'Status', 'Supported Since'])
            sorted_versions = sorted(obj.VERSIONS.items())
            for version, support_status in sorted_versions:
                for support in support_status:
                    cells = [version]
                    sorted_support = sorted(support.items(), reverse=True)
                    cells += [x[1] for x in sorted_support]
                    self._create_table_row(cells, version_tbody)

            # create applicable profile types
            profile_type_description = ('This policy is designed to handle '
                                        'the following profile types:')
            profile_type_section = self._create_section(
                section, 'profile_types', title='Applicable Profile Types')
            field = nodes.line('', profile_type_description)
            profile_type_section.append(field)
            for profile_type in obj.PROFILE_TYPE:
                profile_type_section += self._create_list_item(profile_type)

            # create actions handled
            policy_trigger_description = ('This policy is triggered by the '
                                          'following actions during the '
                                          'respective phases:')
            target_tbody = self._build_table(
                section, 'Policy Triggers',
                ['Action', 'Phase'],
                policy_trigger_description
            )
            sorted_targets = sorted(obj.TARGET, key=lambda tup: tup[1])
            for phase, action in sorted_targets:
                cells = [action, phase]
                self._create_table_row(cells, target_tbody)

            # build properties
            properties_section = self._create_section(section, 'properties',
                                                      title='Properties')
        else:
            properties_section = content

        sorted_schema = sorted(obj.properties_schema.items(),
                               key=cmp_to_key(self._sort_by_type))
        for k, v in sorted_schema:
            self._build_properties(k, v, properties_section)

        # we return the result
        return content

    def _create_section(self, parent, sectionid, title=None, term=None):
        """Create a new section

        :returns: If term is specified, returns a definition node contained
        within the newly created section.  Otherwise return the newly created
        section node.
        """

        idb = nodes.make_id(sectionid)
        section = nodes.section(ids=[idb])
        parent.append(section)

        if term:
            if term != '**':
                section.append(nodes.term('', term))

            definition = nodes.definition()
            section.append(definition)

            return definition

        if title:
            section.append(nodes.title('', title))

        return section

    def _create_list_item(self, str):
        """Creates a new list item

        :returns: List item node
        """
        para = nodes.paragraph()
        para += nodes.strong('', str)

        item = nodes.list_item()
        item += para

        return item

    def _create_def_list(self, parent):
        """Creates a definition list

        :returns: Definition list node
        """

        definition_list = nodes.definition_list()
        parent.append(definition_list)

        return definition_list

    def _sort_by_type(self, x, y):
        """Sort two keys so that map and list types are ordered last."""

        x_key, x_value = x
        y_key, y_value = y

        # if both values are map or list, sort by their keys
        if ((isinstance(x_value, schema.Map) or
             isinstance(x_value, schema.List)) and
                (isinstance(y_value, schema.Map) or
                 isinstance(y_value, schema.List))):
            return (x_key > y_key) - (x_key < y_key)

        # show simple types before maps or list
        if (isinstance(x_value, schema.Map) or
                isinstance(x_value, schema.List)):
            return 1

        if (isinstance(y_value, schema.Map) or
                isinstance(y_value, schema.List)):
            return -1

        return (x_key > y_key) - (x_key < y_key)

    def _create_table_row(self, cells, parent):
        """Creates a table row for cell in cells

        :returns: Row node
        """

        row = nodes.row()
        parent.append(row)

        for c in cells:
            entry = nodes.entry()
            row += entry
            entry += nodes.literal(text=c)

        return row

    def _build_table(self, section, title, headers, description=None):
        """Creates a table with given title, headers and description

        :returns: Table body node
        """

        table_section = self._create_section(section, title, title=title)

        if description:
            field = nodes.line('', description)
            table_section.append(field)

        table = nodes.table()
        tgroup = nodes.tgroup(len(headers))
        table += tgroup

        table_section.append(table)

        for _ in headers:
            tgroup.append(nodes.colspec(colwidth=1))

        # create header
        thead = nodes.thead()
        tgroup += thead
        self._create_table_row(headers, thead)

        tbody = nodes.tbody()
        tgroup += tbody

        # create body consisting of targets
        tbody = nodes.tbody()
        tgroup += tbody

        return tbody

    def _build_properties(self, k, v, definition):
        """Build schema property documentation

        :returns: None
        """

        if isinstance(v, schema.Map):
            newdef = self._create_section(definition, k, term=k)

            if v.schema is None:
                # if it's a map for arbritary values, only include description
                field = nodes.line('', v.description)
                newdef.append(field)
                return

            newdeflist = self._create_def_list(newdef)

            sorted_schema = sorted(v.schema.items(),
                                   key=cmp_to_key(self._sort_by_type))
            for key, value in sorted_schema:
                self._build_properties(key, value, newdeflist)
        elif isinstance(v, schema.List):
            newdef = self._create_section(definition, k, term=k)

            # identify next section as list properties
            field = nodes.line()
            emph = nodes.emphasis('', 'List properties:')
            field.append(emph)
            newdef.append(field)

            newdeflist = self._create_def_list(newdef)

            self._build_properties('**', v.schema['*'], newdeflist)
        else:
            newdef = self._create_section(definition, k, term=k)
            if 'description' in v:
                field = nodes.line('', v['description'])
                newdef.append(field)
            else:
                field = nodes.line('', '++')
                newdef.append(field)


class SchemaProperties(SchemaDirective):
    properties_only = True


class SchemaSpec(SchemaDirective):
    section_title = 'Spec'
    properties_only = False


def setup(app):
    app.add_directive('schemaprops', SchemaProperties)
    app.add_directive('schemaspec', SchemaSpec)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg

from senlin.common.i18n import _

NOTIFICATION_GROUP = cfg.OptGroup(
    name='notification',
)

NOTIFICATION_OPTS = [
    cfg.IntOpt('max_message_size', default=65535,
               help=_('The max size(bytes) of message can be posted to '
                      'notification queue.')),
    cfg.IntOpt('ttl', default=300,
               help=_('The ttl in seconds of a message posted to '
                      'notification queue.')),
]


def register_opts(conf):
    conf.register_group(NOTIFICATION_GROUP)
    conf.register_opts(NOTIFICATION_OPTS, group=NOTIFICATION_GROUP)


def list_opts():
    return {
        NOTIFICATION_GROUP: NOTIFICATION_OPTS,
    }



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from senlin.common import context
from senlin.engine.notifications import heat_endpoint
from senlin import objects
from senlin.tests.unit.common import base


@mock.patch('oslo_messaging.NotificationFilter')
class TestHeatNotificationEndpoint(base.SenlinTestCase):
    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_init(self, mock_rpc, mock_filter):
        x_filter = mock_filter.return_value
        event_map = {
            'orchestration.stack.delete.end': 'DELETE',
        }
        recover_action = {'operation': 'REBUILD'}
        endpoint = heat_endpoint.HeatNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )

        mock_filter.assert_called_once_with(
            publisher_id='^orchestration.*',
            event_type='^orchestration\.stack\..*',
            context={'project_id': '^PROJECT$'})
        mock_rpc.assert_called_once_with()
        self.assertEqual(x_filter, endpoint.filter_rule)
        self.assertEqual(mock_rpc.return_value, endpoint.rpc)
        for e in event_map:
            self.assertIn(e, endpoint.STACK_FAILURE_EVENTS)
            self.assertEqual(event_map[e], endpoint.STACK_FAILURE_EVENTS[e])
        self.assertEqual('PROJECT', endpoint.project_id)
        self.assertEqual('CLUSTER_ID', endpoint.cluster_id)

    @mock.patch.object(context.RequestContext, 'from_dict')
    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info(self, mock_rpc, mock_context, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = heat_endpoint.HeatNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {
            'tags': {
                'cluster_id=CLUSTER_ID',
                'cluster_node_id=FAKE_NODE',
                'cluster_node_index=123',
            },
            'stack_identity': 'PHYSICAL_ID',
            'user_identity': 'USER',
            'state': 'DELETE_COMPLETE',
        }
        metadata = {'timestamp': 'TIMESTAMP'}
        call_ctx = mock.Mock()
        mock_context.return_value = call_ctx

        res = endpoint.info(ctx, 'PUBLISHER', 'orchestration.stack.delete.end',
                            payload, metadata)

        self.assertIsNone(res)
        x_rpc.call.assert_called_once_with(call_ctx, 'node_recover', mock.ANY)
        req = x_rpc.call.call_args[0][2]
        self.assertIsInstance(req, objects.NodeRecoverRequest)
        self.assertEqual('FAKE_NODE', req.identity)
        expected_params = {
            'event': 'DELETE',
            'state': 'DELETE_COMPLETE',
            'stack_id': 'PHYSICAL_ID',
            'timestamp': 'TIMESTAMP',
            'publisher': 'PUBLISHER',
            'operation': 'REBUILD',
        }
        self.assertEqual(expected_params, req.params)

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_event_type_not_interested(self, mock_rpc, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = heat_endpoint.HeatNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {'tags': {'cluster_id': 'CLUSTER_ID'}}
        metadata = {'timestamp': 'TIMESTAMP'}

        res = endpoint.info(ctx, 'PUBLISHER',
                            'orchestration.stack.create.start',
                            payload, metadata)

        self.assertIsNone(res)
        self.assertEqual(0, x_rpc.node_recover.call_count)

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_no_tag(self, mock_rpc, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = heat_endpoint.HeatNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {'tags': None}
        metadata = {'timestamp': 'TIMESTAMP'}

        res = endpoint.info(ctx, 'PUBLISHER', 'orchestration.stack.delete.end',
                            payload, metadata)

        self.assertIsNone(res)
        self.assertEqual(0, x_rpc.node_recover.call_count)

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_empty_tag(self, mock_rpc, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = heat_endpoint.HeatNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {'tags': []}
        metadata = {'timestamp': 'TIMESTAMP'}

        res = endpoint.info(ctx, 'PUBLISHER', 'orchestration.stack.delete.end',
                            payload, metadata)

        self.assertIsNone(res)
        self.assertEqual(0, x_rpc.node_recover.call_count)

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_no_cluster_in_tag(self, mock_rpc, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = heat_endpoint.HeatNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {'tags': ['foo', 'bar']}
        metadata = {'timestamp': 'TIMESTAMP'}

        res = endpoint.info(ctx, 'PUBLISHER', 'orchestration.stack.delete.end',
                            payload, metadata)

        self.assertIsNone(res)
        self.assertEqual(0, x_rpc.node_recover.call_count)

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_no_node_in_tag(self, mock_rpc, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = heat_endpoint.HeatNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {'tags': ['cluster_id=C1ID']}
        metadata = {'timestamp': 'TIMESTAMP'}

        res = endpoint.info(ctx, 'PUBLISHER', 'orchestration.stack.delete.end',
                            payload, metadata)

        self.assertIsNone(res)
        self.assertEqual(0, x_rpc.node_recover.call_count)

    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_cluster_id_not_match(self, mock_rpc, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = heat_endpoint.HeatNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {
            'tags': ['cluster_id=FOOBAR', 'cluster_node_id=N2'],
            'user_identity': 'USER',
        }
        metadata = {'timestamp': 'TIMESTAMP'}

        res = endpoint.info(ctx, 'PUBLISHER', 'orchestration.stack.delete.end',
                            payload, metadata)

        self.assertIsNone(res)
        self.assertEqual(0, x_rpc.node_recover.call_count)

    @mock.patch.object(context.RequestContext, 'from_dict')
    @mock.patch('senlin.rpc.client.get_engine_client')
    def test_info_default_values(self, mock_rpc, mock_context, mock_filter):
        x_rpc = mock_rpc.return_value
        recover_action = {'operation': 'REBUILD'}
        endpoint = heat_endpoint.HeatNotificationEndpoint(
            'PROJECT', 'CLUSTER_ID', recover_action
        )
        ctx = mock.Mock()
        payload = {
            'tags': [
                'cluster_id=CLUSTER_ID',
                'cluster_node_id=NODE_ID'
            ],
            'user_identity': 'USER',
        }
        metadata = {'timestamp': 'TIMESTAMP'}
        call_ctx = mock.Mock()
        mock_context.return_value = call_ctx

        res = endpoint.info(ctx, 'PUBLISHER', 'orchestration.stack.delete.end',
                            payload, metadata)

        self.assertIsNone(res)
        x_rpc.call.assert_called_once_with(call_ctx, 'node_recover', mock.ANY)
        req = x_rpc.call.call_args[0][2]
        self.assertIsInstance(req, objects.NodeRecoverRequest)
        self.assertEqual('NODE_ID', req.identity)
        expected_params = {
            'event': 'DELETE',
            'state': 'Unknown',
            'stack_id': 'Unknown',
            'timestamp': 'TIMESTAMP',
            'publisher': 'PUBLISHER',
            'operation': 'REBUILD',
        }
        self.assertEqual(expected_params, req.params)



# Copyright (c) 2018 NEC, Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from oslo_config import cfg
from oslo_upgradecheck import common_checks
from oslo_upgradecheck import upgradecheck

from senlin.common.i18n import _
from senlin.db import api

from sqlalchemy import MetaData, Table, select, column


class Checks(upgradecheck.UpgradeCommands):

    """Upgrade checks for the senlin-status upgrade check command

    Upgrade checks should be added as separate methods in this class
    and added to _upgrade_checks tuple.
    """

    def _check_healthpolicy(self):
        """Check if version 1.0 health policies exists

        Stein introduces health policy version 1.1 which is incompatible with
        health policy version 1.0.  Users are required to delete version 1.0
        health policies before upgrade and recreate them in version 1.1 format
        after upgrading.
        """

        engine = api.get_engine()
        metadata = MetaData(bind=engine)
        policy = Table('policy', metadata, autoload=True)

        healthpolicy_select = (
            select([column('name')])
            .select_from(policy)
            .where(column('type') == 'senlin.policy.health-1.0')
        )
        healthpolicy_rows = engine.execute(healthpolicy_select).fetchall()

        if not healthpolicy_rows:
            return upgradecheck.Result(upgradecheck.Code.SUCCESS)

        healthpolicy_names = [row[0] for row in healthpolicy_rows]
        error_msg = _('The following version 1.0 health policies must be '
                      'deleted before upgrade: \'{}\'. After upgrading, the '
                      'health policies can be recreated in version 1.1 '
                      'format.').format(', '.join(healthpolicy_names))
        return upgradecheck.Result(upgradecheck.Code.FAILURE, error_msg)

    # The format of the check functions is to return an
    # oslo_upgradecheck.upgradecheck.Result
    # object with the appropriate
    # oslo_upgradecheck.upgradecheck.Code and details set.
    # If the check hits warnings or failures then those should be stored
    # in the returned Result's "details" attribute. The
    # summary will be rolled up at the end of the check() method.
    _upgrade_checks = (
        # In the future there should be some real checks added here
        (_('HealthPolicy'), _check_healthpolicy),
        (_('Policy File JSON to YAML Migration'),
         (common_checks.check_policy_json, {'conf': cfg.CONF})),
    )


def main():
    return upgradecheck.main(
        cfg.CONF, project='senlin', upgrade_command=Checks())


if __name__ == '__main__':
    sys.exit(main())
